import argparse
import atexit
import csv
import os
import time

# noinspection PyPackageRequirements
import socketio

# Parse arguments
parser = argparse.ArgumentParser(
    description="Read and record position into " "CSV files",
    prog="python read_to_file.py",
)
parser.add_argument(
    "-i",
    "--ip",
    type=str,
    default="localhost",
    dest="ip",
    help="IP address to listen on",
)
parser.add_argument(
    "-p", "--port", type=int, default=8080, dest="port", help="Port to listen on"
)
parser.add_argument(
    "-m",
    "--manipulators",
    type=str,
    default="1,2",
    dest="manipulators",
    help="Manipulator IDs, separated by commas with no spaces",
)
args = parser.parse_args()


class ManipulatorMeta:
    """Metadata for a manipulator"""

    def __init__(self, man_id: int):
        """
        Construct a ManipulatorMeta object
        :param man_id: Manipulator ID
        """
        # File
        file_exists = os.path.exists(f"data/{man_id}.csv")
        self.file = open(f"data/{man_id}.csv", "a", newline="")
        self.writer = csv.writer(self.file)
        if not file_exists:
            self.writer.writerow(["time (sec)", "x", "y", "z", "depth"])

        # Previous position
        self.prev_pos = [-1, -1, -1, -1]


# Setup data directory
if not os.path.exists("data"):
    os.makedirs("data")

# Extract manipulators
manipulators = None
try:
    manipulators = {
        int(m): ManipulatorMeta(int(m)) for m in args.manipulators.split(",")
    }
except ValueError:
    print("Invalid manipulator IDs")
    exit(1)
except Exception as e:
    print(e)
    exit(1)

# Setup sockets
sio = socketio.Client()
try:
    # noinspection HttpUrlsUsage
    sio.connect(f"http://{args.ip}:{args.port}")
except Exception as e:
    print(f"Invalid IP address or port: {e}")
    exit(1)


def register_manipulators():
    """Register and bypass calibration for all manipulators"""
    for man_id in manipulators.keys():
        sio.emit("register_manipulator", man_id)
        sio.emit("bypass_calibration", man_id)


# Setup operations
running = True
sleep_time = 1 / 120
movement_threshold = 0.1
start_time = time.time()


def record(man_id: int, pos: list[float], error: str):
    if error == "":
        # Get files to work with
        manipulator = manipulators[man_id]
        dest = manipulator.file
        writer = manipulator.writer
        prev_pos = manipulator.prev_pos

        # Check if update needed
        should_write = False
        for prev, cur in zip(prev_pos, pos):
            if abs(prev - cur) > movement_threshold:
                should_write = True
                break

        # Update if needed
        if should_write:
            writer.writerow([time.time() - start_time] + pos)
            dest.flush()
            manipulator.prev_pos = pos
    elif error == "Manipulator not registered":
        sio.emit("register_manipulator", man_id)
        sio.emit("bypass_calibration", man_id)

    # Pull next position
    if running:
        time.sleep(sleep_time)
        sio.emit("get_pos", man_id, callback=record)


@atexit.register
def cleanup(_, __):
    print("Cleaning up")
    global running
    running = False

    for manipulator in manipulators:
        manipulator.file.flush()
        manipulator.file.close_server()

    sio.disconnect()


print("Begin recording")
for manipulator_id in manipulators.keys():
    sio.emit("get_pos", manipulator_id, callback=record)
    sio.emit("get_pos", manipulator_id, callback=record)



import time

import socketio

sio = socketio.Client()
sio.connect("http://localhost:8081")

manipulator_id = "2"

sio.emit("register_manipulator", manipulator_id)
sio.emit("bypass_calibration", manipulator_id)

sio.emit(
    "set_can_write", {"manipulator_id": manipulator_id, "can_write": True, "hours": 1}
)

sio.emit(
    "goto_pos",
    {"manipulator_id": manipulator_id, "pos": [10, 10, 10, 10], "speed": 4},
)

time.sleep(1)

sio.emit(
    "goto_pos", {"manipulator_id": manipulator_id, "pos": [0, 0, 0, 0], "speed": 4}
)



"""New Scale Manipulator class

Handles logic for calling New Scale API functions. Also includes extra logic for safe
function calls, error handling, managing per-manipulator attributes, and returning the
appropriate callback parameters like in :mod:`ephys_link.new_scale_handler`.
"""

import asyncio
import threading

# noinspection PyPackageRequirements
import socketio

# noinspection PyUnresolvedReferences
from NstMotorCtrl import NstCtrlAxis

import ephys_link.common as com
from ephys_link.platform_manipulator import (
    HOURS_TO_SECONDS,
    MM_TO_UM,
    POSITION_POLL_DELAY,
    PlatformManipulator,
)

# Constants
ACCELERATION_MULTIPLIER = 5
CUTOFF_MULTIPLIER = 0.005
AT_TARGET_FLAG = 0x040000


class NewScaleManipulator(PlatformManipulator):
    def __init__(
        self,
        manipulator_id: str,
        x_axis: NstCtrlAxis,
        y_axis: NstCtrlAxis,
        z_axis: NstCtrlAxis,
    ) -> None:
        """Construct a new Manipulator object

        :param manipulator_id: Manipulator ID
        :param x_axis: X axis object
        :param y_axis: Y axis object
        :param z_axis: Z axis object
        """

        super().__init__()
        self._id = manipulator_id
        self._x = x_axis
        self._y = y_axis
        self._z = z_axis
        self._axes = [self._x, self._y, self._z]

        # Set to CL control
        self._x.SetCL_Enable(True)
        self._y.SetCL_Enable(True)
        self._z.SetCL_Enable(True)

    def query_all_axes(self):
        """Query all axes for their position and status"""
        for axis in self._axes:
            axis.QueryPosStatus()

    def get_pos(self) -> com.PositionalOutputData:
        """Get the current position of the manipulator and convert it into mm

        :return: Callback parameters (position in (x, y, z, 0) (or an empty array on
            error) in mm, error message)
        :rtype: :class:`ephys_link.common.PositionalOutputData`
        """
        self.query_all_axes()

        # Get position data and convert from µm to mm
        try:
            position = [
                self._x.CurPosition / MM_TO_UM,
                self._y.CurPosition / MM_TO_UM,
                self._z.CurPosition / MM_TO_UM,
                self._z.CurPosition / MM_TO_UM,
            ]
            com.dprint(f"[SUCCESS]\t Got position of manipulator {self._id}\n")
            return com.PositionalOutputData(position, "")
        except Exception as e:
            print(f"[ERROR]\t\t Getting position of manipulator {self._id}")
            print(f"{e}\n")
            return com.PositionalOutputData([], "Error getting position")

    async def goto_pos(
        self, position: list[float], speed: float
    ) -> com.PositionalOutputData:
        """Move manipulator to position

        :param position: The position to move to in mm
        :type position: list[float]
        :param speed: The speed to move at (in mm/s)
        :type speed: float
        :return: Callback parameters (position in (x, y, z, w) (or an empty array on
            error), error message)
        :rtype: :class:`ephys_link.common.PositionalOutputData`
        """
        # Check if able to write
        if not self._can_write:
            print(f"[ERROR]\t\t Manipulator {self._id} movement " f"canceled")
            return com.PositionalOutputData([], "Manipulator " "movement canceled")

        # Stop current movement
        if self._is_moving:
            for axis in self._axes:
                axis.Stop()
            self._is_moving = False

        try:
            target_position_um = [axis * MM_TO_UM for axis in position]

            # Restrict target position to just z-axis if inside brain
            if self._inside_brain:
                z_axis = target_position_um[2]
                target_position_um = self.get_pos()["position"]
                target_position_um[2] = z_axis

            # Mark movement as started
            self._is_moving = True

            # Send move command
            speed_um = speed * MM_TO_UM
            for i in range(3):
                self._axes[i].SetCL_Speed(
                    speed_um,
                    speed_um * ACCELERATION_MULTIPLIER,
                    speed_um * CUTOFF_MULTIPLIER,
                )
                self._axes[i].MoveAbsolute(target_position_um[i])

            # Check and wait for completion
            self.query_all_axes()
            while (
                not (self._x.CurStatus & AT_TARGET_FLAG)
                or not (self._y.CurStatus & AT_TARGET_FLAG)
                or not (self._z.CurStatus & AT_TARGET_FLAG)
            ):
                await asyncio.sleep(POSITION_POLL_DELAY)
                self.query_all_axes()

            # Get position
            manipulator_final_position = self.get_pos()["position"]

            # Mark movement as finished
            self._is_moving = False

            com.dprint(
                f"[SUCCESS]\t Moved manipulator {self._id} to position"
                f" {manipulator_final_position}\n"
            )
            return com.PositionalOutputData(manipulator_final_position, "")
        except Exception as e:
            print(f"[ERROR]\t\t Moving manipulator {self._id} to position {position}")
            print(f"{e}\n")
            return com.PositionalOutputData([], "Error moving manipulator")

    async def drive_to_depth(
        self, depth: float, speed: int
    ) -> com.DriveToDepthOutputData:
        """Drive the manipulator to a certain depth

        :param depth: The depth to drive to in mm
        :type depth: float
        :param speed: The speed to drive at in mm/s
        :type speed: int
        :return: Callback parameters (depth (or 0 on error), error message)
        :rtype: :class:`ephys_link.common.DriveToDepthOutputData`
        """
        # Check if able to write
        if not self._can_write:
            print(f"[ERROR]\t\t Manipulator {self._id} movement " f"canceled")
            return com.DriveToDepthOutputData(0, "Manipulator movement canceled")

        # Stop current movement
        if self._is_moving:
            for axis in self._axes:
                axis.Stop()
            self._is_moving = False

        try:
            target_depth_um = depth * MM_TO_UM

            # Mark movement as started
            self._is_moving = True

            # Send move command to just z axis
            speed_um = speed * MM_TO_UM
            self._z.SetCL_Speed(
                speed_um,
                speed_um * ACCELERATION_MULTIPLIER,
                speed_um * CUTOFF_MULTIPLIER,
            )
            self._z.MoveAbsolute(target_depth_um)

            # Check and wait for completion
            self._z.QueryPosStatus()
            while not (self._z.CurStatus & AT_TARGET_FLAG):
                await asyncio.sleep(0.1)
                self._z.QueryPosStatus()

            # Get position
            manipulator_final_position = self.get_pos()["position"]

            # Mark movement as finished
            self._is_moving = False

            com.dprint(
                f"[SUCCESS]\t Moved manipulator {self._id} to position"
                f" {manipulator_final_position}\n"
            )
            return com.DriveToDepthOutputData(manipulator_final_position[3], "")
        except Exception as e:
            print(f"[ERROR]\t\t Moving manipulator {self._id} to depth {depth}")
            print(f"{e}\n")
            # Return 0 and error message on failure
            return com.DriveToDepthOutputData(0, "Error driving " "manipulator")

    def calibrate(self) -> bool:
        """Calibrate the manipulator

        :return: None
        """
        return (
            self._x.CalibrateFrequency()
            and self._y.CalibrateFrequency()
            and self._z.CalibrateFrequency()
        )

    def get_calibrated(self) -> bool:
        """Return the calibration state of the manipulator.

        :return: True if the manipulator is calibrated, False otherwise
        :rtype: bool
        """
        return self._calibrated

    def set_calibrated(self) -> None:
        """Set the manipulator to calibrated

        :return: None
        """
        self._calibrated = True

    def set_can_write(
        self, can_write: bool, hours: float, sio: socketio.AsyncServer
    ) -> None:
        """Set if the manipulator can move

        :param can_write: True if the manipulator can move, False otherwise
        :type can_write: bool
        :param hours: The number of hours to allow the manipulator to move (0 =
            forever)
        :type hours: float
        :param sio: SocketIO object from server to emit reset event
        :type sio: :class:`socketio.AsyncServer`
        :return: None
        """
        self._can_write = can_write

        if can_write and hours > 0:
            if self._reset_timer:
                self._reset_timer.cancel()
            self._reset_timer = threading.Timer(
                hours * HOURS_TO_SECONDS, self.reset_can_write, [sio]
            )
            self._reset_timer.start()

    def get_can_write(self) -> bool:
        """Return if the manipulator can move

        :return: True if the manipulator can move, False otherwise
        :rtype: bool
        """
        return self._can_write

    def reset_can_write(self, sio: socketio.AsyncServer) -> None:
        """Reset the :attr:`can_write` flag

        :param sio: SocketIO object from server to emit reset event
        :type sio: :class:`socketio.AsyncServer`
        :return: None
        """
        self._can_write = False
        asyncio.run(sio.emit("write_disabled", self._id))

    def set_inside_brain(self, inside: bool) -> None:
        """Set if the manipulator is inside the brain

        :param inside: True if the manipulator is inside the brain, False otherwise
        :type inside: bool
        :return: None
        """
        self._inside_brain = inside

    def stop(self) -> None:
        """Stop all axes on manipulator

        :returns None
        """
        for axis in self._axes:
            axis.Stop()
        self._can_write = False
        com.dprint(f"[SUCCESS]\t Stopped manipulator {self._id}\n")



"""WebSocket server and communication handler

Manages the WebSocket server and handles connections and events from the client. For
every event, the server does the following:

1. Extract the arguments passed in the event
2. Log that the event was received
3. Call the appropriate function in :mod:`ephys_link.sensapex_handler` with arguments
4. Relay the response from :mod:`ephys_link.sensapex_handler` to the callback function
"""

import importlib
from importlib import metadata
from typing import Any

import socketio
from aiohttp import web
from aiohttp.web_runner import GracefulExit
from pythonnet import load

from ephys_link import common as com
from ephys_link.platform_handler import PlatformHandler

# Setup server
load("netfx")
sio = socketio.AsyncServer()
app = web.Application()
sio.attach(app)
is_connected = False

# Declare platform handler
platform: PlatformHandler

# Is the server running
is_running = False


# Handle connection events


@sio.event
async def connect(sid, _, __) -> bool:
    """Acknowledge connection to the server

    :param sid: Socket session ID
    :type sid: str
    :param _: WSGI formatted dictionary with request info (unused)
    :type _: dict
    :param __: Authentication details (unused)
    :type __: dict
    :return: False on error to refuse connection. None otherwise.
    :rtype: bool
    """
    print(f"[CONNECTION REQUEST]:\t\t {sid}\n")

    global is_connected
    if not is_connected:
        print(f"[CONNECTION GRANTED]:\t\t {sid}\n")
        is_connected = True
    else:
        print(f"[CONNECTION DENIED]:\t\t {sid}: another client is already connected\n")
        return False


@sio.event
async def disconnect(sid) -> None:
    """Acknowledge disconnection from the server

    :param sid: Socket session ID
    :type sid: str
    :return: None
    """
    print(f"[DISCONNECTION]:\t {sid}\n")

    platform.reset()
    global is_connected
    is_connected = False


# Events


@sio.event
async def get_version(_) -> str:
    """Get the version number of the server

    :param _: Socket session ID (unused)
    :type _: str
    :return: Version number as defined in __version__
    :rtype: str
    """
    return metadata.version("ephys_link")


@sio.event
async def get_manipulators(_) -> com.GetManipulatorsOutputData:
    """Get the list of discoverable manipulators

    :param _: Socket session ID (unused)
    :type _: str
    :return: Callback parameters (manipulators, error message)
    :rtype: :class:`ephys_link.common.GetManipulatorsOutputData`
    """
    com.dprint("[EVENT]\t\t Get discoverable manipulators")

    return platform.get_manipulators()


@sio.event
async def register_manipulator(_, manipulator_id: str) -> str:
    """Register a manipulator with the server

    :param _: Socket session ID (unused)
    :type _: str
    :param manipulator_id: ID of the manipulator to register
    :type manipulator_id: str
    :return: Callback parameter (Error message (on error))
    :rtype: str
    """
    com.dprint(f"[EVENT]\t\t Register manipulator: {manipulator_id}")

    return platform.register_manipulator(manipulator_id)


@sio.event
async def unregister_manipulator(_, manipulator_id: str) -> str:
    """Unregister a manipulator from the server

    :param _: Socket session ID (unused)
    :type _: str
    :param manipulator_id: ID of the manipulator to unregister
    :type manipulator_id: str
    :return: Callback parameter (Error message (on error))
    :rtype: str
    """
    com.dprint(f"[EVENT]\t\t Unregister manipulator: {manipulator_id}")

    return platform.unregister_manipulator(manipulator_id)


@sio.event
async def get_pos(_, manipulator_id: str) -> com.PositionalOutputData:
    """Position of manipulator request

    :param _: Socket session ID (unused)
    :type _: str
    :param manipulator_id: ID of manipulator to pull position from
    :type manipulator_id: str
    :return: Callback parameters (manipulator ID, position in (x, y, z, w) (or an empty
        array on error) in mm, error message)
    :rtype: :class:`ephys_link.common.PositionalOutputData`
    """
    # com.dprint(f"[EVENT]\t\t Get position of manipulator" f" {manipulator_id}")

    return platform.get_pos(manipulator_id)


@sio.event
async def get_angles(_, manipulator_id: str) -> com.AngularOutputData:
    """Angles of manipulator request

    :param _: Socket session ID (unused)
    :type _: str
    :param manipulator_id: ID of manipulator to pull angles from
    :type manipulator_id: str
    :return: Callback parameters (manipulator ID, angles in (yaw, pitch, roll) (or an empty
        array on error) in degrees, error message)
    :rtype: :class:`ephys_link.common.AngularOutputData`
    """

    return platform.get_angles(manipulator_id)


@sio.event
async def goto_pos(
    _, data: com.GotoPositionInputDataFormat
) -> com.PositionalOutputData:
    """Move manipulator to position

    :param _: Socket session ID (unused)
    :type _: str
    :param data: Data containing manipulator ID, position in mm, and speed in mm/s
    :type data: :class:`ephys_link.common.GotoPositionInputDataFormat`
    :return: Callback parameters (manipulator ID, position in (x, y, z, w) (or an empty
        tuple on error) in mm, error message)
    :rtype: :class:`ephys_link.common.PositionalOutputData`
    """
    try:
        manipulator_id = data["manipulator_id"]
        pos = data["pos"]
        speed = data["speed"]

    except KeyError:
        manipulator_id = data["manipulator_id"] if "manipulator_id" in data else -1
        print(f"[ERROR]\t\t Invalid data for manipulator {manipulator_id}\n")
        return com.PositionalOutputData([], "Invalid data format")

    except Exception as e:
        print(f"[ERROR]\t\t Error in goto_pos: {e}\n")
        return com.PositionalOutputData([], "Error in goto_pos")

    com.dprint(f"[EVENT]\t\t Move manipulator {manipulator_id} " f"to position {pos}")

    return await platform.goto_pos(manipulator_id, pos, speed)


@sio.event
async def drive_to_depth(
    _, data: com.DriveToDepthInputDataFormat
) -> com.DriveToDepthOutputData:
    """Drive to depth

    :param _: Socket session ID (unused)
    :type _: str
    :param data: Data containing manipulator ID, depth in mm, and speed in mm/s
    :type data: :class:`ephys_link.common.DriveToDepthInputDataFormat`
    :return: Callback parameters (manipulator ID, depth (or -1 on error) in mm, error message
        )
    :rtype: :class:`ephys_link.common.DriveToDepthOutputData`
    """
    try:
        manipulator_id = data["manipulator_id"]
        depth = data["depth"]
        speed = data["speed"]

    except KeyError:
        manipulator_id = data["manipulator_id"] if "manipulator_id" in data else -1
        print(f"[ERROR]\t\t Invalid data for manipulator {manipulator_id}\n")
        return com.DriveToDepthOutputData(-1, "Invalid data " "format")

    except Exception as e:
        print(f"[ERROR]\t\t Error in drive_to_depth: {e}\n")
        return com.DriveToDepthOutputData(-1, "Error in drive_to_depth")

    com.dprint(f"[EVENT]\t\t Drive manipulator {manipulator_id} " f"to depth {depth}")

    return await platform.drive_to_depth(manipulator_id, depth, speed)


@sio.event
async def set_inside_brain(
    _, data: com.InsideBrainInputDataFormat
) -> com.StateOutputData:
    """Set the inside brain state

    :param _: Socket session ID (unused)
    :type _: str
    :param data: Data containing manipulator ID and inside brain state
    :type data: :class:`ephys_link.common.InsideBrainInputDataFormat`
    :return: Callback parameters (manipulator ID, inside, error message)
    :rtype: :class:`ephys_link.common.StateOutputData`
    """
    try:
        manipulator_id = data["manipulator_id"]
        inside = data["inside"]

    except KeyError:
        manipulator_id = data["manipulator_id"] if "manipulator_id" in data else -1
        print(f"[ERROR]\t\t Invalid data for manipulator {manipulator_id}\n")
        return com.StateOutputData(False, "Invalid data format")
    except Exception as e:
        print(f"[ERROR]\t\t Error in inside_brain: {e}\n")
        return com.StateOutputData(False, "Error in set_inside_brain")

    com.dprint(
        f"[EVENT]\t\t Set manipulator {manipulator_id} inside brain to "
        f'{"true" if inside else "false"}'
    )

    return platform.set_inside_brain(manipulator_id, inside)


@sio.event
async def calibrate(_, manipulator_id: str) -> str:
    """Calibrate manipulator

    :param _: Socket session ID (unused)
    :type _: str
    :param manipulator_id: ID of manipulator to calibrate
    :type manipulator_id: str
    :return: Callback parameters (manipulator ID, error message)
    :rtype: str
    """
    com.dprint(f"[EVENT]\t\t Calibrate manipulator" f" {manipulator_id}")

    return await platform.calibrate(manipulator_id, sio)


@sio.event
async def bypass_calibration(_, manipulator_id: str) -> str:
    """Bypass calibration of manipulator

    :param _: Socket session ID (unused)
    :type _: str
    :param manipulator_id: ID of manipulator to bypass calibration
    :type manipulator_id: str
    :return: Callback parameters (manipulator ID, error message)
    :rtype: str
    """
    com.dprint(f"[EVENT]\t\t Bypass calibration of manipulator" f" {manipulator_id}")

    return platform.bypass_calibration(manipulator_id)


@sio.event
async def set_can_write(_, data: com.CanWriteInputDataFormat) -> com.StateOutputData:
    """Set manipulator can_write state

    :param _: Socket session ID (unused)
    :type _: str
    :param data: Data containing manipulator ID and can_write brain state
    :type data: :class:`ephys_link.common.CanWriteInputDataFormat`
    :return: Callback parameters (manipulator ID, can_write, error message)
    :rtype: :class:`ephys_link.common.StateOutputData`
    """
    try:
        manipulator_id = data["manipulator_id"]
        can_write = data["can_write"]
        hours = data["hours"]

    except KeyError:
        manipulator_id = data["manipulator_id"] if "manipulator_id" in data else -1
        print(f"[ERROR]\t\t Invalid data for manipulator {manipulator_id}\n")
        return com.StateOutputData(False, "Invalid data " "format")

    except Exception as e:
        print(f"[ERROR]\t\t Error in inside_brain: {e}\n")
        return com.StateOutputData(False, "Error in set_can_write")

    com.dprint(
        f"[EVENT]\t\t Set manipulator {manipulator_id} can_write state to "
        f'{"true" if can_write else "false"}'
    )

    return platform.set_can_write(manipulator_id, can_write, hours, sio)


@sio.event
def stop(_) -> bool:
    """Stop all manipulators

    :param _: Socket session ID (unused)
    :type _: str
    :return: True if successful, False otherwise
    :rtype: bool
    """
    com.dprint("[EVENT]\t\t Stop all manipulators")

    return platform.stop()


@sio.on("*")
async def catch_all(_, __, data: Any) -> None:
    """Catch all event

    :param _: Socket session ID (unused)
    :type _: str
    :param __: Client ID (unused)
    :type __: str
    :param data: Data received from client
    :type data: Any
    :return: None
    """
    print(f"[UNKNOWN EVENT]:\t {data}")


# Handle server start and end


def launch_server(platform_type: str, server_port: int, pathfinder_port: int) -> None:
    """Launch the server

    :param platform_type: Parsed argument for platform type
    :type platform_type: str
    :param server_port: HTTP port to serve the server
    :type server_port: int
    :param pathfinder_port: Port New Scale Pathfinder's server is on
    :type pathfinder_port: int
    :return: None
    """

    # Import correct manipulator handler
    global platform
    if platform_type == "sensapex":
        platform = importlib.import_module(
            "ephys_link.platforms.sensapex_handler"
        ).SensapexHandler()
    elif platform_type == "new_scale":
        platform = importlib.import_module(
            "ephys_link.platforms.new_scale_handler"
        ).NewScaleHandler()
    elif platform_type == "new_scale_pathfinder":
        platform = importlib.import_module(
            "ephys_link.platforms.new_scale_pathfinder_handler"
        ).NewScalePathfinderHandler(pathfinder_port)
    else:
        exit(f"[ERROR]\t\t Invalid manipulator type: {platform_type}")

    # Preamble
    print(f"=== Ephys Link v{metadata.version('ephys_link')} ===")

    # List available manipulators
    print("Available Manipulators:")
    print(platform.get_manipulators()["manipulators"])

    print("\n(Shutdown server with CTRL+Pause/Break)\n")

    # Mark that server is running
    global is_running
    is_running = True
    web.run_app(app, port=server_port)


def close_server(_, __) -> None:
    """Close the server"""
    print("[INFO]\t\t Closing server")

    # Stop movement
    platform.stop()  # noqa

    # Exit
    raise GracefulExit()



from unittest import TestCase
from unittest.mock import Mock

# noinspection PyPackageRequirements
import socketio

from src.ephys_link import PositionalOutputData


class GetPosTestCase(TestCase):
    """Tests get_pos event"""

    def setUp(self):
        """Setup test case"""
        self.sio = socketio.Client()
        self.mock = Mock()

        self.sio.connect("http://localhost:8081")

    def test_get_pos_unregistered(self):
        """Test get_pos event with unregistered manipulator"""
        self.sio.emit("get_pos", "1", callback=self.mock)
        self.sio.emit("bypass_calibration", "1")
        self.wait_for_callback()

        self.mock.assert_called_with(
            PositionalOutputData([], "Manipulator not registered")
        )

    def test_get_pos_registered(self):
        """Test get_pos event with registered manipulator"""
        self.sio.emit("register_manipulator", "1")
        self.sio.emit("bypass_calibration", "1")
        self.sio.emit("get_pos", "1", callback=self.mock)
        self.wait_for_callback()

        args = self.mock.call_args.args[0]
        self.assertEqual(len(args["position"]), 4)
        self.assertEqual(args["error"], "")

    def tearDown(self) -> None:
        """Cleanup test case"""
        self.sio.disconnect()

    def wait_for_callback(self):
        """Wait for callback to be called"""
        while not self.mock.called:
            pass



from unittest import TestCase
from unittest.mock import Mock

# noinspection PyPackageRequirements
import socketio


class RegisterManipulatorTestCase(TestCase):
    """Tests manipulator registration"""

    def setUp(self) -> None:
        """Setup test case"""
        self.sio = socketio.Client()
        self.mock = Mock()

        self.sio.connect("http://localhost:8081")

    def test_register_manipulator(self):
        """Tests registering a manipulator"""
        self.sio.emit("register_manipulator", "1", callback=self.mock)
        self.wait_for_callback()

        self.mock.assert_called_with("")

    def test_re_register_manipulator(self):
        """Test re-registering a manipulator"""
        self.sio.emit("register_manipulator", 1)
        self.sio.emit("register_manipulator", 1, callback=self.mock)
        self.wait_for_callback()

        self.mock.assert_called_with("Manipulator already registered")

    def test_register_unknown_manipulator(self):
        """Test registering an unknown manipulator"""
        self.sio.emit("register_manipulator", 99, callback=self.mock)
        self.wait_for_callback()

        self.mock.assert_called_with("Manipulator not found")

    def tearDown(self) -> None:
        """Cleanup test case"""
        self.sio.disconnect()

    def wait_for_callback(self):
        """Wait for callback to be called"""
        while not self.mock.called:
            pass



from unittest import TestCase
from unittest.mock import Mock

# noinspection PyPackageRequirements
import socketio


# noinspection DuplicatedCode
class StopTest(TestCase):
    """Tests stop event"""

    DRIVE_SPEED = 8000

    def setUp(self) -> None:
        """Setup test case"""
        self.sio = socketio.Client()
        self.mock = Mock()

        self.sio.connect("http://localhost:8080")

    def test_stop_event(self):
        """Test stopping movement with event"""
        self.sio.emit("register_manipulator", 1)
        self.sio.emit("register_manipulator", 2)
        self.sio.emit("bypass_calibration", 1)
        self.sio.emit("bypass_calibration", 2)
        self.sio.emit(
            "set_can_write", {"manipulator_id": 1, "can_write": True, "hours": 1}
        )
        self.sio.emit(
            "set_can_write", {"manipulator_id": 2, "can_write": True, "hours": 1}
        )

        self.sio.emit(
            "goto_pos",
            {"manipulator_id": 1, "pos": [0, 0, 0, 0], "speed": self.DRIVE_SPEED},
            callback=self.mock,
        )
        self.sio.emit(
            "goto_pos",
            {"manipulator_id": 2, "pos": [0, 0, 0, 0], "speed": self.DRIVE_SPEED},
            callback=self.mock,
        )

        self.sio.emit(
            "goto_pos",
            {
                "manipulator_id": 1,
                "pos": [10000, 10000, 10000, 10000],
                "speed": self.DRIVE_SPEED,
            },
        )
        self.sio.emit(
            "goto_pos",
            {
                "manipulator_id": 2,
                "pos": [10000, 10000, 10000, 10000],
                "speed": self.DRIVE_SPEED,
            },
        )

        while self.mock.call_count != 2:
            pass

        self.sio.sleep(1)

        # Test stop
        self.mock.called = False
        self.sio.emit("stop", callback=self.mock)
        self.wait_for_callback()
        stop_arg = self.mock.call_args.args[0]

        # Test no calibration
        self.sio.emit(
            "goto_pos",
            {"manipulator_id": 1, "pos": [0, 0, 0, 0], "speed": self.DRIVE_SPEED},
            callback=self.mock,
        )
        self.wait_for_callback()
        args = self.mock.call_args.args[0]

        # Bring back to home
        self.mock.call_count = 0
        self.sio.emit("bypass_calibration", 1)
        self.sio.emit("bypass_calibration", 2)
        self.sio.emit(
            "set_can_write", {"manipulator_id": 1, "can_write": True, "hours": 1}
        )
        self.sio.emit(
            "set_can_write", {"manipulator_id": 2, "can_write": True, "hours": 1}
        )
        self.sio.emit(
            "goto_pos",
            {
                "manipulator_id": 1,
                "pos": [10000, 10000, 10000, 10000],
                "speed": self.DRIVE_SPEED,
            },
            callback=self.mock,
        )
        self.sio.emit(
            "goto_pos",
            {
                "manipulator_id": 2,
                "pos": [10000, 10000, 10000, 10000],
                "speed": self.DRIVE_SPEED,
            },
            callback=self.mock,
        )

        while self.mock.call_count != 2:
            pass

        # Asserts
        self.assertTrue(stop_arg)
        self.assertEqual(len(args), 2)
        self.assertEqual(args["error"], "Cannot write to manipulator")

    def tearDown(self) -> None:
        """Cleanup test case"""
        self.sio.disconnect()

    def wait_for_callback(self):
        """Wait for callback to be called"""
        while not self.mock.called:
            pass
        self.mock.called = False



"""Defines what the properties and required functionality of a manipulator are.

Most functionality will be implemented on the platform handler side. This is mostly
for enforcing implementation of the stop method and hold common properties.
"""
from abc import ABC, abstractmethod

# Constants
MM_TO_UM = 1000
HOURS_TO_SECONDS = 3600
POSITION_POLL_DELAY = 0.1


class PlatformManipulator(ABC):
    """An abstract class that defines the interface for a manipulator."""

    def __init__(self):
        """Initialize manipulator."""

        self._id = None
        self._calibrated = False
        self._inside_brain = False
        self._can_write = False
        self._reset_timer = None
        self._is_moving = False

    @abstractmethod
    def stop(self) -> None:
        """Stop all axes on manipulator

        :returns None
        """
        raise NotImplementedError



"""Handle communications with platform specific API

Handles relaying WebSocket messages to the appropriate platform API functions and
conducting error checks on the input and output values

Function names here are the same as the WebSocket events. They are called when the
server receives an event from a client. In general, each function does the following:

1. Receive extracted arguments from :mod:`ephys_link.server`
2. Call and check the appropriate platform API function (overloaded by each platform)
3. Log/handle successes and failures
4. Return the callback parameters to :mod:`ephys_link.server`
"""

from abc import ABC, abstractmethod

# noinspection PyPackageRequirements
import socketio

from ephys_link import common as com


class PlatformHandler(ABC):
    """An abstract class that defines the interface for a manipulator handler."""

    def __init__(self):
        """Initialize the manipulator handler with a dictionary of manipulators."""

        # Registered manipulators are stored as a dictionary of IDs (string) to
        # manipulator objects
        self.manipulators = {}
        self.type = "sensapex"  # Remove this after #165

    # Platform Handler Methods

    def reset(self) -> bool:
        """Reset handler

        :return: True if successful, False otherwise
        :rtype: bool
        """
        stop_result = self.stop()
        self.manipulators.clear()
        return stop_result

    def stop(self) -> bool:
        """Stop handler

        :return: True if successful, False otherwise
        :rtype: bool
        """
        try:
            for manipulator in self.manipulators.values():
                if hasattr(manipulator, "stop"):
                    manipulator.stop()
            return True
        except Exception as e:
            print(f"[ERROR]\t\t Could not stop manipulators: {e}\n")
            return False

    def get_manipulators(self) -> com.GetManipulatorsOutputData:
        """Get all registered manipulators

        :return: Callback parameters (manipulators, error)
        :rtype: :class:`ephys_link.common.GetManipulatorsOutputData`
        """
        devices = []
        error = "Error getting manipulators"
        try:
            # noinspection PyUnresolvedReferences
            devices = self._get_manipulators()
            error = ""
        except Exception as e:
            print(f"[ERROR]\t\t Getting manipulators: {type(e)}: {e}\n")
        finally:
            return com.GetManipulatorsOutputData(devices, self.type, error)

    def register_manipulator(self, manipulator_id: str) -> str:
        """Register a manipulator

        :param manipulator_id: The ID of the manipulator to register.
        :type manipulator_id: str
        :return: Callback parameter (Error message (on error))
        :rtype: str
        """
        # Check if manipulator is already registered
        if manipulator_id in self.manipulators:
            print(f"[ERROR]\t\t Manipulator already registered:" f" {manipulator_id}\n")
            return "Manipulator already registered"

        try:
            # Register manipulator
            self._register_manipulator(manipulator_id)
            com.dprint(f"[SUCCESS]\t Registered manipulator: {manipulator_id}\n")
            return ""

        except ValueError as ve:
            # Manipulator not found in UMP
            print(f"[ERROR]\t\t Manipulator not found: {manipulator_id}: {ve}\n")
            return "Manipulator not found"

        except Exception as e:
            # Other error
            print(f"[ERROR]\t\t Registering manipulator: {manipulator_id}")
            print(f"{type(e)}: {e}\n")
            return "Error registering manipulator"

    def unregister_manipulator(self, manipulator_id: str) -> str:
        """Unregister a manipulator

        :param manipulator_id: The ID of the manipulator to unregister.
        :type manipulator_id: str
        :return: Callback parameters (error message (on error))
        """
        # Check if manipulator is not registered
        if manipulator_id not in self.manipulators:
            print(f"[ERROR]\t\t Manipulator not registered: {manipulator_id}\n")
            return "Manipulator not registered"

        try:
            # Unregister manipulator
            self._unregister_manipulator(manipulator_id)

            com.dprint(f"[SUCCESS]\t Unregistered manipulator: {manipulator_id}\n")
            return ""
        except Exception as e:
            # Other error
            print(f"[ERROR]\t\t Unregistering manipulator: {manipulator_id}")
            print(f"{e}\n")
            return "Error unregistering manipulator"

    def get_pos(self, manipulator_id: str) -> com.PositionalOutputData:
        """Get the current position of a manipulator

        :param manipulator_id: The ID of the manipulator to get the position of.
        :type manipulator_id: str
        :return: Callback parameters (manipulator ID, position in (x, y, z, w) (or an
            empty array on error) in mm, error message)
        :rtype: :class:`ephys_link.common.PositionalOutputData`
        """
        try:
            # Check calibration status
            if (
                hasattr(self.manipulators[manipulator_id], "get_calibrated")
                and not self.manipulators[manipulator_id].get_calibrated()
            ):
                print(f"[ERROR]\t\t Calibration not complete: {manipulator_id}\n")
                return com.PositionalOutputData([], "Manipulator not calibrated")

            # Get position
            return self._get_pos(manipulator_id)

        except KeyError:
            # Manipulator not found in registered manipulators
            print(f"[ERROR]\t\t Manipulator not registered: {manipulator_id}")
            return com.PositionalOutputData([], "Manipulator not registered")

    def get_angles(self, manipulator_id: str) -> com.AngularOutputData:
        """Get the current position of a manipulator

        :param manipulator_id: The ID of the manipulator to get the position of.
        :type manipulator_id: str
        :return: Callback parameters (manipulator ID, angles in (yaw, pitch, roll) (or an
            empty array on error) in degrees, error message)
        :rtype: :class:`ephys_link.common.AngularOutputData`
        """
        try:
            # Check calibration status
            if (
                hasattr(self.manipulators[manipulator_id], "get_calibrated")
                and not self.manipulators[manipulator_id].get_calibrated()
            ):
                print(f"[ERROR]\t\t Calibration not complete: {manipulator_id}\n")
                return com.AngularOutputData([], "Manipulator not calibrated")

            # Get position
            return self._get_angles(manipulator_id)

        except KeyError:
            # Manipulator not found in registered manipulators
            print(f"[ERROR]\t\t Manipulator not registered: {manipulator_id}")
            return com.AngularOutputData([], "Manipulator not registered")

    async def goto_pos(
        self, manipulator_id: str, position: list[float], speed: int
    ) -> com.PositionalOutputData:
        """Move manipulator to position

        :param manipulator_id: The ID of the manipulator to move
        :type manipulator_id: str
        :param position: The position to move to in (x, y, z, w) in mm
        :type position: list[float]
        :param speed: The speed to move at (in mm/s)
        :type speed: int
        :return: Callback parameters (manipulator ID, position in (x, y, z, w) (or an
                 empty array on error) in mm, error message)
        :rtype: :class:`ephys_link.common.PositionalOutputData`
        """
        try:
            # Check calibration status
            if not self.manipulators[manipulator_id].get_calibrated():
                print(f"[ERROR]\t\t Calibration not complete: {manipulator_id}\n")
                return com.PositionalOutputData([], "Manipulator not calibrated")

            # Check write state
            if not self.manipulators[manipulator_id].get_can_write():
                print(f"[ERROR]\t\t Cannot write to manipulator: {manipulator_id}")
                return com.PositionalOutputData([], "Cannot write to manipulator")

            return await self._goto_pos(manipulator_id, position, speed)

        except KeyError:
            # Manipulator not found in registered manipulators
            print(f"[ERROR]\t\t Manipulator not registered: {manipulator_id}\n")
            return com.PositionalOutputData([], "Manipulator not registered")

    async def drive_to_depth(
        self, manipulator_id: str, depth: float, speed: int
    ) -> com.DriveToDepthOutputData:
        """Drive manipulator to depth

        :param manipulator_id: The ID of the manipulator to drive
        :type manipulator_id: str
        :param depth: The depth to drive to in mm
        :type depth: float
        :param speed: The speed to drive at (in mm/s)
        :type speed: int
        :return: Callback parameters (manipulator ID, depth (or 0 on error) in mm, error
                 message)
        :rtype: :class:`ephys_link.common.DriveToDepthOutputData`
        """
        try:
            # Check calibration status
            if not self.manipulators[manipulator_id].get_calibrated():
                print(f"[ERROR]\t\t Calibration not complete: {manipulator_id}\n")
                return com.DriveToDepthOutputData(0, "Manipulator not calibrated")

            # Check write state
            if not self.manipulators[manipulator_id].get_can_write():
                print(f"[ERROR]\t\t Cannot write to manipulator: {manipulator_id}")
                return com.DriveToDepthOutputData(0, "Cannot write to manipulator")

            return await self._drive_to_depth(manipulator_id, depth, speed)

        except KeyError:
            # Manipulator not found in registered manipulators
            print(f"[ERROR]\t\t Manipulator not registered: {manipulator_id}\n")
            return com.DriveToDepthOutputData(0, "Manipulator " "not registered")

    def set_inside_brain(
        self, manipulator_id: str, inside: bool
    ) -> com.StateOutputData:
        """Set manipulator inside brain state (restricts motion)

        :param manipulator_id: The ID of the manipulator to set the state of
        :type manipulator_id: str
        :param inside: True if inside brain, False if outside
        :type inside: bool
        :return: Callback parameters (manipulator ID, inside, error message)
        :rtype: :class:`ephys_link.common.StateOutputData`
        """
        try:
            # Check calibration status
            if (
                hasattr(self.manipulators[manipulator_id], "get_calibrated")
                and not self.manipulators[manipulator_id].get_calibrated()
            ):
                print("[ERROR]\t\t Calibration not complete\n")
                return com.StateOutputData(False, "Manipulator not calibrated")

            return self._set_inside_brain(manipulator_id, inside)

        except KeyError:
            # Manipulator not found in registered manipulators
            print(f"[ERROR]\t\t Manipulator {manipulator_id} not registered\n")
            return com.StateOutputData(False, "Manipulator not " "registered")

        except Exception as e:
            # Other error
            print(
                f"[ERROR]\t\t Set manipulator {manipulator_id} inside brain " f"state"
            )
            print(f"{e}\n")
            return com.StateOutputData(False, "Error setting " "inside brain")

    async def calibrate(self, manipulator_id: str, sio: socketio.AsyncServer) -> str:
        """Calibrate manipulator

        :param manipulator_id: ID of manipulator to calibrate
        :type manipulator_id: str
        :param sio: SocketIO object (to call sleep)
        :type sio: :class:`socketio.AsyncServer`
        :return: Callback parameters (manipulator ID, error message)
        :rtype: str
        """
        try:
            # Check write state
            if not self.manipulators[manipulator_id].get_can_write():
                print(f"[ERROR]\t\t Cannot write to manipulator: {manipulator_id}")
                return "Cannot write to manipulator"

            return await self._calibrate(manipulator_id, sio)

        except KeyError:
            # Manipulator not found in registered manipulators
            print(f"[ERROR]\t\t Manipulator {manipulator_id} not registered\n")
            return "Manipulator not registered"

        except Exception as e:
            # Other error
            print(f"[ERROR]\t\t Calibrate manipulator {manipulator_id}")
            print(f"{e}\n")
            return "Error calibrating manipulator"

    def bypass_calibration(self, manipulator_id: str) -> str:
        """Bypass calibration of manipulator

        :param manipulator_id: ID of manipulator to bypass calibration
        :type manipulator_id: str
        :return: Callback parameters (manipulator ID, error message)
        :rtype: str
        """
        try:
            # Bypass calibration
            return self._bypass_calibration(manipulator_id)

        except KeyError:
            # Manipulator not found in registered manipulators
            print(f"[ERROR]\t\t Manipulator {manipulator_id} not registered\n")
            return "Manipulator not registered"

        except Exception as e:
            # Other error
            print(f"[ERROR]\t\t Bypass calibration of manipulator {manipulator_id}")
            print(f"{e}\n")
            return "Error bypassing calibration"

    def set_can_write(
        self,
        manipulator_id: str,
        can_write: bool,
        hours: float,
        sio: socketio.AsyncServer,
    ) -> com.StateOutputData:
        """Set manipulator can_write state (enables/disabled moving manipulator)

        :param manipulator_id: The ID of the manipulator to set the state of
        :type manipulator_id: str
        :param can_write: True if allowed to move, False if outside
        :type can_write: bool
        :param hours: The number of hours to allow writing (0 = forever)
        :type hours: float
        :param sio: SocketIO object from server to emit reset event
        :type sio: :class:`socketio.AsyncServer`
        :return: Callback parameters (manipulator ID, can_write, error message)
        :rtype: :class:`ephys_link.common.StateOutputData`
        """
        try:
            return self._set_can_write(manipulator_id, can_write, hours, sio)
        except KeyError:
            # Manipulator not found in registered manipulators
            print(f"[ERROR]\t\t Manipulator not registered: {manipulator_id}\n")
            return com.StateOutputData(False, "Manipulator not " "registered")

        except Exception as e:
            # Other error
            print(f"[ERROR]\t\t Set manipulator {manipulator_id} can_write state")
            print(f"{e}\n")
            return com.StateOutputData(False, "Error setting " "can_write")

    # Platform specific methods to override

    @abstractmethod
    def _get_manipulators(self) -> list:
        """Get all registered manipulators

        :return: List of manipulator IDs
        :rtype: list
        """
        pass

    @abstractmethod
    def _register_manipulator(self, manipulator_id: str) -> None:
        """Register a manipulator

        :param manipulator_id: The ID of the manipulator to register.
        :type manipulator_id: str
        :return: None
        """
        pass

    @abstractmethod
    def _unregister_manipulator(self, manipulator_id: str) -> None:
        """Unregister a manipulator

        :param manipulator_id: The ID of the manipulator to unregister.
        :type manipulator_id: str
        :return: None
        """
        pass

    @abstractmethod
    def _get_pos(self, manipulator_id: str) -> com.PositionalOutputData:
        """Get the current position of a manipulator

        :param manipulator_id: The ID of the manipulator to get the position of.
        :type manipulator_id: int
        :return: Callback parameters (manipulator ID, position in (x, y, z, w) (or an
            empty array on error) in mm, error message)
        :rtype: :class:`ephys_link.common.PositionalOutputData`
        """
        pass

    @abstractmethod
    def _get_angles(self, manipulator_id: str) -> com.AngularOutputData:
        """Get the current angles of a manipulator

        :param manipulator_id: The ID of the manipulator to get the position of.
        :type manipulator_id: int
        :return: Callback parameters (manipulator ID, position in (yaw, pitch, roll) (or an
            empty array on error) in degrees, error message)
        :rtype: :class:`ephys_link.common.AngularOutputData`
        """
        pass

    @abstractmethod
    async def _goto_pos(
        self, manipulator_id: str, position: list[float], speed: int
    ) -> com.PositionalOutputData:
        """Move manipulator to position

        :param manipulator_id: The ID of the manipulator to move
        :type manipulator_id: str
        :param position: The position to move to in mm
        :type position: list[float]
        :param speed: The speed to move at (in mm/s)
        :type speed: int
        :return: Callback parameters (manipulator ID, position in (x, y, z, w) (or an
                 empty array on error) in mm, error message)
        :rtype: :class:`ephys_link.common.PositionalOutputData`
        """
        pass

    @abstractmethod
    async def _drive_to_depth(
        self, manipulator_id: str, depth: float, speed: int
    ) -> com.DriveToDepthOutputData:
        """Drive manipulator to depth

        :param manipulator_id: The ID of the manipulator to drive
        :type manipulator_id: str
        :param depth: The depth to drive to in mm
        :type depth: float
        :param speed: The speed to drive at (in mm/s)
        :type speed: int
        :return: Callback parameters (manipulator ID, depth (or 0 on error), error
                 message)
        :rtype: :class:`ephys_link.common.DriveToDepthOutputData`
        """
        pass

    @abstractmethod
    def _set_inside_brain(
        self, manipulator_id: str, inside: bool
    ) -> com.StateOutputData:
        """Set manipulator inside brain state (restricts motion)

        :param manipulator_id: The ID of the manipulator to set the state of
        :type manipulator_id: str
        :param inside: True if inside brain, False if outside
        :type inside: bool
        :return: Callback parameters (manipulator ID, inside, error message)
        :rtype: :class:`ephys_link.common.StateOutputData`
        """
        pass

    @abstractmethod
    async def _calibrate(self, manipulator_id: str, sio: socketio.AsyncServer) -> str:
        """Calibrate manipulator

        :param manipulator_id: ID of manipulator to calibrate
        :type manipulator_id: str
        :param sio: SocketIO object (to call sleep)
        :type sio: :class:`socketio.AsyncServer`
        :return: Callback parameters (manipulator ID, error message)
        :rtype: str
        """
        pass

    @abstractmethod
    def _bypass_calibration(self, manipulator_id: str) -> str:
        """Bypass calibration of manipulator

        :param manipulator_id: ID of manipulator to bypass calibration
        :type manipulator_id: str
        :return: Callback parameters (manipulator ID, error message)
        :rtype: str
        """
        pass

    @abstractmethod
    def _set_can_write(
        self,
        manipulator_id: str,
        can_write: bool,
        hours: float,
        sio: socketio.AsyncServer,
    ) -> com.StateOutputData:
        """Set manipulator can_write state (enables/disabled moving manipulator)

        :param manipulator_id: The ID of the manipulator to set the state of
        :type manipulator_id: str
        :param can_write: True if allowed to move, False if outside
        :type can_write: bool
        :param hours: The number of hours to allow writing (0 = forever)
        :type hours: float
        :param sio: SocketIO object from server to emit reset event
        :type sio: :class:`socketio.AsyncServer`
        :return: Callback parameters (manipulator ID, can_write, error message)
        :rtype: :class:`ephys_link.common.StateOutputData`
        """
        pass



"""Handle communications with New Scale's HTTP server

Implements New Scale specific API calls.

This is a subclass of :class:`ephys_link.platform_handler.PlatformHandler`.
"""

from json import loads
from urllib import request

# noinspection PyPackageRequirements
import socketio

from ephys_link import common as com
from ephys_link.platform_handler import PlatformHandler


class NewScalePathfinderHandler(PlatformHandler):
    """Handler for New Scale HTTP server"""

    # Valid New Scale manipulator IDs
    VALID_MANIPULATOR_IDS = {
        "A",
        "B",
        "C",
        "D",
        "E",
        "F",
        "G",
        "H",
        "I",
        "J",
        "K",
        "L",
        "M",
        "N",
        "O",
        "P",
        "Q",
        "R",
        "S",
        "T",
        "U",
        "V",
        "W",
        "X",
        "Y",
        "Z",
        "AA",
        "AB",
        "AC",
        "AD",
        "AE",
        "AF",
        "AG",
        "AH",
        "AI",
        "AJ",
        "AK",
        "AL",
        "AM",
        "AN",
    }

    def __init__(self, port: int = 8080) -> None:
        """
        Initialize New Scale via Pathfinder handler

        :param port: Port of New Scale Pathfinder HTTP server
        :type port: int
        """
        super().__init__()

        self.type = "new_scale_pathfinder"

        self.port = port

        # Test connection to New Scale HTTP server
        try:
            request.urlopen(f"http://localhost:{self.port}")
        except Exception as e:
            raise ValueError(
                f"New Scale HTTP server not online on port {self.port}"
            ) from e

    def query_data(self) -> dict:
        """Query New Scale HTTP server for data and return as dict

        :return: dict of data (originally in JSON)
        """
        try:
            return loads(request.urlopen(f"http://localhost:{self.port}").read())
        except Exception as e:
            print(f"[ERROR]\t\t Unable to query for New Scale data: {type(e)} {e}\n")

    def query_manipulator_data(self, manipulator_id: str) -> dict:
        """Query New Scale HTTP server for data on a specific manipulator

        :param manipulator_id: manipulator ID
        :return: dict of data (originally in JSON)
        :raises ValueError: if manipulator ID is not found in query
        """
        data_query = self.query_data()["ProbeArray"]
        manipulator_data = data_query[self.manipulators[manipulator_id]]

        # If the order of the manipulators switched (somehow)
        if manipulator_data["Id"] != manipulator_id:
            # Recalculate index and get data
            (manipulator_index, manipulator_data) = next(
                (
                    (index, data)
                    for index, data in enumerate(self.query_data()["ProbeArray"])
                    if data["Id"] == manipulator_id
                ),
                (None, None),
            )
            # Update index in manipulators dict
            if manipulator_index:
                self.manipulators[manipulator_id] = manipulator_index

        # If data query was unsuccessful
        if not manipulator_data:
            raise ValueError(f"Unable to find manipulator {manipulator_id}")

        # Return data
        return manipulator_data

    def _get_manipulators(self) -> list:
        return [probe["Id"] for probe in self.query_data()["ProbeArray"]]

    def _register_manipulator(self, manipulator_id: str) -> None:
        # Check if ID is a valid New Scale manipulator ID
        if manipulator_id not in self.VALID_MANIPULATOR_IDS:
            raise ValueError(f"Invalid manipulator ID {manipulator_id}")

        # Check if ID is connected
        if manipulator_id not in self._get_manipulators():
            raise ValueError(f"Manipulator {manipulator_id} not connected")

        # Get index of the manipulator
        manipulator_index = next(
            (
                index
                for index, data in enumerate(self.query_data()["ProbeArray"])
                if data["Id"] == manipulator_id
            ),
            None,
        )
        if manipulator_index is None:
            raise ValueError(f"Unable to find manipulator {manipulator_id}")
        self.manipulators[manipulator_id] = manipulator_index

    def _unregister_manipulator(self, manipulator_id: str) -> None:
        del self.manipulators[manipulator_id]

    def _get_pos(self, manipulator_id: str) -> com.PositionalOutputData:
        """Get the current position of the manipulator in mm

        :param manipulator_id: manipulator ID
        :return: Callback parameters (position in (x, y, z, w) (or an empty array on
            error) in mm, error message)
        """
        manipulator_data = self.query_manipulator_data(manipulator_id)

        return com.PositionalOutputData(
            [
                manipulator_data["Tip_X_ML"],
                manipulator_data["Tip_Y_AP"],
                manipulator_data["Tip_Z_DV"],
                0,
            ],
            "",
        )

    def _get_angles(self, manipulator_id: str) -> com.AngularOutputData:
        """Get the current angles of the manipulator in degrees

        :param manipulator_id: manipulator ID
        :return: Callback parameters (angles in (yaw, pitch, roll) (or an empty array on
            error) in degrees, error message)
        """
        manipulator_data = self.query_manipulator_data(manipulator_id)

        return com.AngularOutputData(
            [
                manipulator_data["Polar"],
                manipulator_data["Pitch"],
                manipulator_data["Roll"],
            ],
            "",
        )

    async def _goto_pos(
        self, manipulator_id: str, position: list[float], speed: int
    ) -> com.PositionalOutputData:
        pass

    async def _drive_to_depth(
        self, manipulator_id: str, depth: float, speed: int
    ) -> com.DriveToDepthOutputData:
        pass

    def _set_inside_brain(
        self, manipulator_id: str, inside: bool
    ) -> com.StateOutputData:
        pass

    async def _calibrate(self, manipulator_id: str, sio: socketio.AsyncServer) -> str:
        pass

    def _bypass_calibration(self, manipulator_id: str) -> str:
        return ""

    def _set_can_write(
        self,
        manipulator_id: str,
        can_write: bool,
        hours: float,
        sio: socketio.AsyncServer,
    ) -> com.StateOutputData:
        pass



from unittest import TestCase
from unittest.mock import Mock

# noinspection PyPackageRequirements
import socketio

from src.ephys_link import DriveToDepthOutputData


# noinspection DuplicatedCode
class DriveToDepthTest(TestCase):
    """Tests drive to depth event"""

    DRIVE_SPEED = 5000

    def setUp(self) -> None:
        """Setup test case"""
        self.sio = socketio.Client()
        self.mock = Mock()

        self.sio.connect("http://localhost:8080")

    def test_move_no_asserts(self):
        """Test movement without asserts"""
        self.sio.emit("register_manipulator", 1)
        self.sio.emit("register_manipulator", 2)
        self.sio.emit("bypass_calibration", 1)
        self.sio.emit("bypass_calibration", 2)
        self.sio.emit(
            "set_can_write", {"manipulator_id": 1, "can_write": True, "hours": 1}
        )
        self.sio.emit(
            "set_can_write", {"manipulator_id": 2, "can_write": True, "hours": 1}
        )

        self.sio.emit(
            "drive_to_depth",
            {"manipulator_id": 1, "depth": 0, "speed": self.DRIVE_SPEED},
            callback=self.mock,
        )
        self.sio.emit(
            "drive_to_depth",
            {"manipulator_id": 2, "depth": 0, "speed": self.DRIVE_SPEED},
            callback=self.mock,
        )

        self.sio.emit(
            "drive_to_depth",
            {"manipulator_id": 1, "depth": 10000, "speed": self.DRIVE_SPEED},
            callback=self.mock,
        )
        self.sio.emit(
            "drive_to_depth",
            {"manipulator_id": 2, "depth": 10000, "speed": self.DRIVE_SPEED},
            callback=self.mock,
        )

        while self.mock.call_count != 4:
            pass

    def test_move_with_asserts(self):
        """Test movement with asserts"""
        self.sio.emit("register_manipulator", 1)
        self.sio.emit("bypass_calibration", 1)
        self.sio.emit(
            "set_can_write", {"manipulator_id": 1, "can_write": True, "hours": 1}
        )

        self.sio.emit(
            "drive_to_depth",
            {"manipulator_id": 1, "depth": 0, "speed": self.DRIVE_SPEED},
            callback=self.mock,
        )
        self.wait_for_callback()
        self.assertEqual(self.mock.call_args.args[0]["error"], "")

        self.sio.emit(
            "drive_to_depth",
            {"manipulator_id": 1, "depth": 10000, "speed": self.DRIVE_SPEED},
            callback=self.mock,
        )
        self.wait_for_callback()

    def test_move_unregistered(self):
        """Test movement with unregistered manipulator"""
        self.sio.emit(
            "drive_to_depth",
            {"manipulator_id": 1, "depth": 0, "speed": self.DRIVE_SPEED},
            callback=self.mock,
        )
        self.wait_for_callback()
        self.mock.assert_called_with(
            DriveToDepthOutputData(0, "Manipulator not registered")
        )

    def tearDown(self) -> None:
        """Cleanup test case"""
        self.sio.disconnect()

    def wait_for_callback(self):
        """Wait for callback to be called"""
        while not self.mock.called:
            pass
        self.mock.called = False



"""Commonly used functions and dictionaries

Contains globally used helper functions and typed dictionaries (to be used as
callback parameters)
"""

from typing import TypedDict

# Debugging flag
DEBUG = False


def set_debug(debug: bool) -> None:
    """Set debug flag

    :param debug: True to enable debug mode, False to disable
    :type debug: bool
    :return: None
    """
    global DEBUG
    DEBUG = debug


def dprint(message: str) -> None:
    """Print message if debug is enabled

    :param message: Message to print
    :type message: str
    :return: None
    """
    if DEBUG:
        print(message)


# Input data formats
class GotoPositionInputDataFormat(TypedDict):
    """Data format for :func:`server.goto_pos`"""

    manipulator_id: str
    pos: list[float]
    speed: int


class InsideBrainInputDataFormat(TypedDict):
    """Data format for :func:`server.set_inside_brain`"""

    manipulator_id: str
    inside: bool


class DriveToDepthInputDataFormat(TypedDict):
    """Data format for :func:`server.drive_to_depth`"""

    manipulator_id: str
    depth: float
    speed: int


class CanWriteInputDataFormat(TypedDict):
    """Data format for :func:`server.set_can_write`"""

    manipulator_id: str
    can_write: bool
    hours: float


# Output data dictionaries
class GetManipulatorsOutputData(dict):
    """Output format for (manipulators)

    :param manipulators: Tuple of manipulator IDs (as strings)
    :type manipulators: list
    :param manipulator_type: Type of the output data (temporary solution until #165 is implemented)
    :type manipulator_type: str
    :param error: Error message
    :type error: str

    :example: Example generated dictionary
        :code:`{"manipulators": ["1", "2"], "error": ""}`
    """

    def __init__(self, manipulators: list, manipulator_type: str, error: str) -> None:
        """Constructor"""
        super(GetManipulatorsOutputData, self).__init__(
            manipulators=manipulators, type=manipulator_type, error=error
        )


class PositionalOutputData(dict):
    """Output format for (position, error)

    :param position: Position in mm (as a tuple, can be empty) in X, Y, Z, W order
    :type position: list
    :param error: Error message
    :type error: str

    :example: Example generated dictionary
        :code:`{"position": [10.429, 12.332, 2.131, 12.312], "error": ""}`
    """

    def __init__(self, position: list, error: str) -> None:
        """Constructor"""
        super(PositionalOutputData, self).__init__(position=position, error=error)


class AngularOutputData(dict):
    """Output format for (angles, error)

    :param angles: Angles in degrees (as a tuple, can be empty) in yaw, pitch, roll order
    :type angles: list
    :param error: Error message
    :type error: str
    """

    def __init__(self, angles: list, error: str) -> None:
        """Constructor"""
        super(AngularOutputData, self).__init__(angles=angles, error=error)


class DriveToDepthOutputData(dict):
    """Output format for depth driving (depth, error)

    :param depth: Depth in mm
    :type depth: float
    :param error: Error message
    :type error: str

    :example: Example generated dictionary :code:`{"depth": 0.123, "error": ""}`
    """

    def __init__(self, depth: float, error: str) -> None:
        """Create drive to depth output data dictionary"""
        super(DriveToDepthOutputData, self).__init__(depth=depth, error=error)


class StateOutputData(dict):
    """Output format for (state, error)

    :param state: State of the event
    :type state: bool
    :param error: Error message
    :type error: str

    :example: Example generated dictionary :code:`{"state": True, "error": ""}`
    """

    def __init__(self, state: bool, error: str) -> None:
        """Create state output data dictionary"""
        super(StateOutputData, self).__init__(state=state, error=error)



import argparse
import signal
import time
from importlib import metadata
from threading import Event, Thread

import serial
import serial.tools.list_ports as ports

from ephys_link import common as com
from ephys_link import server

# Setup Arduino serial port (emergency stop)
poll_rate = 0.05
kill_serial_event = Event()
poll_serial_thread: Thread


def poll_serial(kill_event: Event, serial_port: str) -> None:
    """Continuously poll serial port for data

    :param kill_event: Event to stop polling
    :type kill_event: Event
    :param serial_port: The serial port to poll
    :type serial_port: str
    :return: None
    """
    target_port = serial_port
    if serial_port is None:
        # Search for serial ports
        for port, desc, _ in ports.comports():
            if "Arduino" in desc or "USB Serial Device" in desc:
                target_port = port
                break
    elif serial_port == "no-e-stop":
        # Stop polling if no-e-stop is specified
        return None

    ser = serial.Serial(target_port, 9600, timeout=poll_rate)
    while not kill_event.is_set():
        if ser.in_waiting > 0:
            ser.readline()
            # Cause a break
            com.dprint("[EMERGENCY STOP]\t\t Stopping all manipulators")
            server.platform.stop()
            ser.reset_input_buffer()
        time.sleep(poll_rate)
    print("Close poll")
    ser.close()


def close_serial(_, __) -> None:
    """Close the serial connection"""
    print("[INFO]\t\t Closing serial")
    kill_serial_event.set()
    poll_serial_thread.join()


# Setup argument parser
parser = argparse.ArgumentParser(
    description="Electrophysiology Manipulator Link: a websocket interface for"
    " manipulators in electrophysiology experiments",
    prog="python -m ephys-link",
)
# parser.add_argument("-g", "--gui", dest="gui", action="store_true", help="Launches GUI")
parser.add_argument(
    "-t",
    "--type",
    type=str,
    dest="type",
    default="sensapex",
    help='Manipulator type (i.e. "sensapex", "new_scale", or "new_scale_pathfinder").'
    ' Default: "sensapex"',
)
parser.add_argument(
    "-d", "--debug", dest="debug", action="store_true", help="Enable debug mode"
)
parser.add_argument(
    "-p",
    "--port",
    type=int,
    default=8081,
    dest="port",
    help="Port to serve on. Default: 8081 (avoids conflict with other HTTP servers)",
)
parser.add_argument(
    "--pathfinder_port",
    type=int,
    default=8080,
    dest="pathfinder_port",
    help="Port New Scale Pathfinder's server is on. Default: 8080",
)
parser.add_argument(
    "-s",
    "--serial",
    type=str,
    default="no-e-stop",
    dest="serial",
    nargs="?",
    help="Emergency stop serial port (i.e. COM3). Default: disables emergency stop",
)
parser.add_argument(
    "-v",
    "--version",
    action="version",
    version=f"Electrophysiology Manipulator Link v{metadata.version('ephys_link')}",
    help="Print version and exit",
)


def main() -> None:
    """Main function"""

    # Parse arguments
    args = parser.parse_args()
    com.set_debug(args.debug)

    # Setup serial port
    if args.serial != "no-e-stop":
        # Register serial exit
        signal.signal(signal.SIGTERM, close_serial)
        signal.signal(signal.SIGINT, close_serial)

        # Start emergency stop system if serial is provided
        global poll_serial_thread
        poll_serial_thread = Thread(
            target=poll_serial,
            args=(
                kill_serial_event,
                args.serial,
            ),
            daemon=True,
        )
        poll_serial_thread.start()

    # Register server exit
    signal.signal(signal.SIGTERM, server.close_server)
    signal.signal(signal.SIGINT, server.close_server)

    # Launch with parsed arguments on main thread
    server.launch_server(args.type, args.port, args.pathfinder_port)


if __name__ == "__main__":
    main()



from unittest import TestCase
from unittest.mock import Mock

# noinspection PyPackageRequirements
import socketio


class InsideBrainTestCase(TestCase):
    """Test movement inside brain event"""

    DRIVE_SPEED = 8000

    def setUp(self):
        """Setup test case"""
        self.sio = socketio.Client()
        self.mock = Mock()

        self.sio.connect("http://localhost:8080")

    def test_move_inside_brain(self):
        """Test goto_pos event with manipulator inside the brain"""
        self.sio.emit("register_manipulator", 1)
        self.sio.emit("bypass_calibration", 1)
        self.sio.emit(
            "set_can_write", {"manipulator_id": 1, "can_write": True, "hours": 1}
        )

        self.sio.emit(
            "goto_pos",
            {"manipulator_id": 1, "pos": (0, 0, 0, 0), "speed": self.DRIVE_SPEED},
            callback=self.mock,
        )
        self.wait_for_callback()

        self.sio.emit(
            "set_inside_brain",
            {"manipulator_id": 1, "inside": True},
            callback=self.mock,
        )
        self.wait_for_callback()

        self.sio.emit(
            "goto_pos",
            {
                "manipulator_id": 1,
                "pos": (10000, 10000, 10000, 50),
                "speed": self.DRIVE_SPEED,
            },
            callback=self.mock,
        )
        self.wait_for_callback()

        self.sio.emit("get_pos", 1, callback=self.mock)
        args = self.mock.call_args.args[0]
        self.mock.called = False

        self.sio.emit(
            "set_inside_brain",
            {"manipulator_id": 1, "inside": False},
            callback=self.mock,
        )
        self.wait_for_callback()
        self.sio.emit(
            "goto_pos",
            {
                "manipulator_id": 1,
                "pos": (10000, 10000, 10000, 10000),
                "speed": self.DRIVE_SPEED,
            },
            callback=self.mock,
        )
        self.wait_for_callback()

        self.assertLess(abs(args["position"][0]), 1)
        self.assertLess(abs(args["position"][1]), 1)
        self.assertLess(abs(args["position"][2]), 1)

    def tearDown(self) -> None:
        """Cleanup test case"""
        self.sio.disconnect()

    def wait_for_callback(self):
        """Wait for callback to be called"""
        while not self.mock.called:
            pass
        self.mock.called = False



"""Handle communications with Sensapex uMp API

Implements Sensapex uMp specific API calls including coordinating the usage of the
:class:`ephys_link.platforms.sensapex_manipulator.SensapexManipulator` class.

This is a subclass of :class:`ephys_link.platform_handler.PlatformHandler`.
"""
from pathlib import Path

# noinspection PyPackageRequirements
import socketio
from sensapex import UMP, UMError

import ephys_link.common as com
from ephys_link.platform_handler import PlatformHandler
from ephys_link.platforms.sensapex_manipulator import SensapexManipulator


class SensapexHandler(PlatformHandler):
    """Handler for Sensapex platform"""

    def __init__(self):
        super().__init__()

        # Establish connection to Sensapex API (exit if connection fails)
        UMP.set_library_path(
            str(Path(__file__).parent.parent.absolute()) + "/resources/"
        )
        self.ump = UMP.get_ump()
        if self.ump is None:
            raise ValueError("Unable to connect to uMp")

    def _get_manipulators(self) -> list:
        return list(map(str, self.ump.list_devices()))

    def _register_manipulator(self, manipulator_id: str) -> None:
        if not manipulator_id.isnumeric():
            raise ValueError("Manipulator ID must be numeric")

        self.manipulators[manipulator_id] = SensapexManipulator(
            self.ump.get_device(int(manipulator_id))
        )

    def _unregister_manipulator(self, manipulator_id: str) -> None:
        del self.manipulators[manipulator_id]

    def _get_pos(self, manipulator_id: str) -> com.PositionalOutputData:
        return self.manipulators[manipulator_id].get_pos()

    def _get_angles(self, manipulator_id: str) -> com.AngularOutputData:
        raise NotImplementedError

    async def _goto_pos(
        self, manipulator_id: str, position: list[float], speed: int
    ) -> com.PositionalOutputData:
        return await self.manipulators[manipulator_id].goto_pos(position, speed)

    async def _drive_to_depth(
        self, manipulator_id: str, depth: float, speed: int
    ) -> com.DriveToDepthOutputData:
        return await self.manipulators[manipulator_id].drive_to_depth(depth, speed)

    def _set_inside_brain(
        self, manipulator_id: str, inside: bool
    ) -> com.StateOutputData:
        self.manipulators[manipulator_id].set_inside_brain(inside)
        com.dprint(
            f"[SUCCESS]\t Set inside brain state for manipulator:"
            f" {manipulator_id}\n"
        )
        return com.StateOutputData(inside, "")

    async def _calibrate(self, manipulator_id: str, sio: socketio.AsyncServer) -> str:
        try:
            # Move manipulator to max position
            await self.manipulators[manipulator_id].goto_pos(
                [20000, 20000, 20000, 20000], 2000
            )

            # Call calibrate
            self.manipulators[manipulator_id].call_calibrate()

            # Wait for calibration to complete
            still_working = True
            while still_working:
                cur_pos = self.manipulators[manipulator_id].get_pos()["position"]

                # Check difference between current and target position
                for prev, cur in zip([10000, 10000, 10000, 10000], cur_pos):
                    if abs(prev - cur) > 1:
                        still_working = True
                        break
                    still_working = False

                # Sleep for a bit
                await sio.sleep(0.5)

            # Calibration complete
            self.manipulators[manipulator_id].set_calibrated()
            com.dprint(f"[SUCCESS]\t Calibrated manipulator {manipulator_id}\n")
            return ""
        except UMError as e:
            # SDK call error
            print(f"[ERROR]\t\t Calling calibrate manipulator {manipulator_id}")
            print(f"{e}\n")
            return "Error calling calibrate"

    def _bypass_calibration(self, manipulator_id: str) -> str:
        self.manipulators[manipulator_id].set_calibrated()
        com.dprint(
            f"[SUCCESS]\t Bypassed calibration for manipulator" f" {manipulator_id}\n"
        )
        return ""

    def _set_can_write(
        self,
        manipulator_id: str,
        can_write: bool,
        hours: float,
        sio: socketio.AsyncServer,
    ) -> com.StateOutputData:
        self.manipulators[manipulator_id].set_can_write(can_write, hours, sio)
        com.dprint(
            f"[SUCCESS]\t Set can_write state for manipulator" f" {manipulator_id}\n"
        )
        return com.StateOutputData(can_write, "")



import threading
from unittest import TestCase
from unittest.mock import Mock

# noinspection PyPackageRequirements
import socketio

from src.ephys_link import PositionalOutputData


# noinspection DuplicatedCode
class MoveTest(TestCase):
    """Tests movement event"""

    DRIVE_SPEED = 5000

    def setUp(self) -> None:
        """Setup test case"""
        self.sio = socketio.Client()
        self.mock = Mock()

        self.sio.connect("http://localhost:8081")

    def test_move_no_asserts(self):
        """Test movement without asserts"""
        # self.sio.emit("register_manipulator", "0")
        # self.sio.emit("bypass_calibration", "0")
        # self.sio.emit(
        #     "set_can_write", {"manipulator_id": "0", "can_write": True, "hours": 1}
        # )
        # self.sio.emit("register_manipulator", "1")
        # self.sio.emit("bypass_calibration", "1")
        # self.sio.emit(
        #     "set_can_write", {"manipulator_id": "1", "can_write": True, "hours": 1}
        # )
        # self.sio.emit("register_manipulator", "2")
        # self.sio.emit("bypass_calibration", "2")
        # self.sio.emit(
        #     "set_can_write", {"manipulator_id": "2", "can_write": True, "hours": 1}
        # )
        self.sio.emit("register_manipulator", "4")
        self.sio.emit("bypass_calibration", "4")
        self.sio.emit(
            "set_can_write", {"manipulator_id": "4", "can_write": True, "hours": 1}
        )

        # Call get_pos() continuously
        def get_pos():
            pos_mock = Mock()
            while True:
                self.sio.emit("get_pos", "4", callback=pos_mock)
                while not pos_mock.called:
                    self.sio.sleep(0.1)

        threading.Thread(target=get_pos, daemon=True).start()

        # self.sio.emit(
        #     "goto_pos",
        #     {"manipulator_id": "0", "pos": [0, 0, 0, 0], "speed": self.DRIVE_SPEED},
        #     callback=self.mock,
        # )
        # self.sio.emit(
        #     "goto_pos",
        #     {"manipulator_id": "1", "pos": [0, 0, 0, 0], "speed": self.DRIVE_SPEED},
        #     callback=self.mock,
        # )
        # self.sio.emit(
        #     "goto_pos",
        #     {"manipulator_id": "2", "pos": [0, 0, 0, 0], "speed": self.DRIVE_SPEED},
        #     callback=self.mock,
        # )
        self.sio.emit(
            "goto_pos",
            {"manipulator_id": "4", "pos": [0, 0, 0, 0], "speed": self.DRIVE_SPEED},
            callback=self.mock,
        )

        # self.sio.emit(
        #     "goto_pos",
        #     {
        #         "manipulator_id": "0",
        #         "pos": [10, 10, 10, 10],
        #         "speed": self.DRIVE_SPEED,
        #     },
        #     callback=self.mock,
        # )
        # self.sio.emit(
        #     "goto_pos",
        #     {
        #         "manipulator_id": "1",
        #         "pos": [10, 10, 10, 10],
        #         "speed": self.DRIVE_SPEED,
        #     },
        #     callback=self.mock,
        # )
        # self.sio.emit(
        #     "goto_pos",
        #     {
        #         "manipulator_id": "2",
        #         "pos": [10000, 10000, 10000, 10000],
        #         "speed": self.DRIVE_SPEED,
        #     },
        #     callback=self.mock,
        # )
        self.sio.emit(
            "goto_pos",
            {
                "manipulator_id": "4",
                "pos": [20, 20, 20, 20],
                "speed": self.DRIVE_SPEED,
            },
            callback=self.mock,
        )

        self.sio.emit(
            "goto_pos",
            {
                "manipulator_id": "4",
                "pos": [10, 10, 10, 10],
                "speed": self.DRIVE_SPEED,
            },
            callback=self.mock,
        )

        while self.mock.call_count != 3:
            pass

    def test_move_with_asserts(self):
        """Test movement with asserts"""
        self.sio.emit("register_manipulator", "0")
        self.sio.emit("register_manipulator", "1")
        self.sio.emit("bypass_calibration", "0")
        self.sio.emit("bypass_calibration", "1")
        # self.sio.emit("register_manipulator", "2")
        # self.sio.emit("bypass_calibration", "2")
        self.sio.emit(
            "set_can_write", {"manipulator_id": "0", "can_write": True, "hours": 1}
        )
        self.sio.emit(
            "set_can_write", {"manipulator_id": "1", "can_write": True, "hours": 1}
        )
        # self.sio.emit(
        #     "set_can_write", {"manipulator_id": "2", "can_write": True, "hours": 1}
        # )

        self.sio.emit(
            "goto_pos",
            {"manipulator_id": "1", "pos": [0, 0, 0, 0], "speed": self.DRIVE_SPEED},
            callback=self.mock,
        )
        self.wait_for_callback()
        args = self.mock.call_args.args[0]
        self.assertEqual(args["error"], "")
        self.assertEqual(len(args["position"]), 4)

        self.sio.emit(
            "goto_pos",
            {
                "manipulator_id": "1",
                "pos": [10, 10, 10, 10],
                "speed": self.DRIVE_SPEED,
            },
            callback=self.mock,
        )
        self.wait_for_callback()

    def test_move_unregistered(self):
        """Test movement with unregistered manipulator"""
        self.sio.emit(
            "goto_pos",
            {"manipulator_id": "1", "pos": [0, 0, 0, 0], "speed": self.DRIVE_SPEED},
            callback=self.mock,
        )
        self.wait_for_callback()
        self.mock.assert_called_with(
            PositionalOutputData([], "Manipulator not registered")
        )

    def test_move_no_write(self):
        """Test movement with no write"""
        self.sio.emit("register_manipulator", "1")
        self.sio.emit("bypass_calibration", "1")
        self.sio.emit(
            "goto_pos",
            {"manipulator_id": "1", "pos": [0, 0, 0, 0], "speed": self.DRIVE_SPEED},
            callback=self.mock,
        )
        self.wait_for_callback()
        self.mock.assert_called_with(
            PositionalOutputData([], "Cannot write to manipulator")
        )

    def test_move_sequence(self):
        """Test insertion driving sequence"""
        self.sio.emit("register_manipulator", "1")
        self.sio.emit("bypass_calibration", "1")
        self.sio.emit(
            "set_can_write", {"manipulator_id": "1", "can_write": True, "hours": 1}
        )

        # Send chain
        self.sio.emit(
            "goto_pos",
            {"manipulator_id": "1", "pos": [0, 0, 0, 0], "speed": self.DRIVE_SPEED},
            callback=self.mock,
        )
        self.sio.emit(
            "goto_pos",
            {"manipulator_id": "1", "pos": [7.5, 0, 0, 0], "speed": self.DRIVE_SPEED},
            callback=self.mock,
        )
        self.sio.emit(
            "goto_pos",
            {"manipulator_id": "1", "pos": [7.5, 7.5, 0, 0], "speed": self.DRIVE_SPEED},
            callback=self.mock,
        )
        self.sio.emit(
            "goto_pos",
            {
                "manipulator_id": "1",
                "pos": [7.5, 7.5, 7.5, 0],
                "speed": self.DRIVE_SPEED,
            },
            callback=self.mock,
        )

        while self.mock.call_count != 4:
            pass

    def tearDown(self) -> None:
        """Cleanup test case"""
        self.sio.disconnect()

    def wait_for_callback(self):
        """Wait for callback to be called"""
        while not self.mock.called:
            pass
        self.mock.called = False



"""Handle communications with New Scale API

Implements New Scale specific API calls.

This is a subclass of :class:`ephys_link.platform_handler.PlatformHandler`.
"""

import importlib

# noinspection PyPackageRequirements
import clr

# noinspection PyPackageRequirements
import socketio

from ephys_link import common as com
from ephys_link.platform_handler import PlatformHandler


class NewScaleHandler(PlatformHandler):
    """Handler for New Scale platform"""

    def __init__(self) -> None:
        """Initialize New Scale handler"""
        super().__init__()

        self.type = "new_scale"

        # Load New Scale API
        # noinspection PyUnresolvedReferences
        clr.AddReference("ephys_link/resources/NstMotorCtrl")
        # noinspection PyUnresolvedReferences
        from NstMotorCtrl import NstCtrlHostIntf

        self.ctrl = NstCtrlHostIntf()

        # Connect manipulators and initialize
        self.ctrl.ShowProperties()
        self.ctrl.Initialize()

    def _get_manipulators(self) -> list:
        return list(map(str, range(self.ctrl.PortCount)))

    def _register_manipulator(self, manipulator_id: str) -> None:
        # Check if ID is numeric
        if not manipulator_id.isnumeric():
            raise ValueError("Manipulator ID must be numeric")

        # Check if ID is connected
        if manipulator_id not in self._get_manipulators():
            raise ValueError(f"Manipulator {manipulator_id} not connected")

        # Check if there are enough axes
        if int(manipulator_id) * 3 + 2 >= self.ctrl.AxisCount:
            raise ValueError(f"Manipulator {manipulator_id} has no axes")

        # Register manipulator
        first_axis_index = int(manipulator_id) * 3
        self.manipulators[manipulator_id] = importlib.import_module(
            "ephys_link.platforms.new_scale_manipulator"
        ).NewScaleManipulator(
            manipulator_id,
            self.ctrl.GetAxis(first_axis_index),
            self.ctrl.GetAxis(first_axis_index + 1),
            self.ctrl.GetAxis(first_axis_index + 2),
        )

    def _unregister_manipulator(self, manipulator_id: str) -> None:
        del self.manipulators[manipulator_id]

    def _get_pos(self, manipulator_id: str) -> com.PositionalOutputData:
        return self.manipulators[manipulator_id].get_pos()

    def _get_angles(self, manipulator_id: str) -> com.AngularOutputData:
        raise NotImplementedError

    async def _goto_pos(
        self, manipulator_id: str, position: list[float], speed: int
    ) -> com.PositionalOutputData:
        return await self.manipulators[manipulator_id].goto_pos(position, speed)

    async def _drive_to_depth(
        self, manipulator_id: str, depth: float, speed: int
    ) -> com.DriveToDepthOutputData:
        return await self.manipulators[manipulator_id].drive_to_depth(depth, speed)

    def _set_inside_brain(
        self, manipulator_id: str, inside: bool
    ) -> com.StateOutputData:
        self.manipulators[manipulator_id].set_inside_brain(inside)
        com.dprint(
            f"[SUCCESS]\t Set inside brain state for manipulator:"
            f" {manipulator_id}\n"
        )
        return com.StateOutputData(inside, "")

    async def _calibrate(self, manipulator_id: str, sio: socketio.AsyncServer) -> str:
        return (
            ""
            if self.manipulators[manipulator_id].calibrate()
            else "Error calling calibrate"
        )

    def _bypass_calibration(self, manipulator_id: str) -> str:
        self.manipulators[manipulator_id].set_calibrated()
        com.dprint(
            f"[SUCCESS]\t Bypassed calibration for manipulator" f" {manipulator_id}\n"
        )
        return ""

    def _set_can_write(
        self,
        manipulator_id: str,
        can_write: bool,
        hours: float,
        sio: socketio.AsyncServer,
    ) -> com.StateOutputData:
        self.manipulators[manipulator_id].set_can_write(can_write, hours, sio)
        com.dprint(
            f"[SUCCESS]\t Set can_write state for manipulator" f" {manipulator_id}\n"
        )
        return com.StateOutputData(can_write, "")



from unittest import TestCase
from unittest.mock import Mock

# noinspection PyPackageRequirements
import socketio

sio = socketio.Client()
sio.connect("http://localhost:8080")


@sio.event
def write_disabled(data):
    print(f"[EVENT]\t\t Write disabled for manipulator {data}")


# noinspection DuplicatedCode
class CanWriteTest(TestCase):
    """Tests can_write event"""

    DRIVE_SPEED = 8000

    def setUp(self) -> None:
        """Setup test case"""
        self.mock = Mock()

    def test_can_write_timer(self):
        """Test movement without asserts"""
        sio.emit("register_manipulator", 1)
        sio.emit("register_manipulator", 2)
        sio.emit("bypass_calibration", 1)
        sio.emit("bypass_calibration", 2)
        sio.emit(
            "set_can_write",
            {"manipulator_id": 1, "can_write": True, "hours": 0.00027777777},
        )
        sio.emit(
            "set_can_write",
            {"manipulator_id": 2, "can_write": True, "hours": 0.00027777777},
        )

        sio.emit(
            "goto_pos",
            {"manipulator_id": 1, "pos": [0, 0, 0, 0], "speed": self.DRIVE_SPEED},
            callback=self.mock,
        )
        sio.emit(
            "goto_pos",
            {"manipulator_id": 2, "pos": [0, 0, 0, 0], "speed": self.DRIVE_SPEED},
            callback=self.mock,
        )

        sio.emit(
            "goto_pos",
            {
                "manipulator_id": 1,
                "pos": [10000, 10000, 10000, 10000],
                "speed": self.DRIVE_SPEED,
            },
            callback=self.mock,
        )
        sio.emit(
            "goto_pos",
            {
                "manipulator_id": 2,
                "pos": [10000, 10000, 10000, 10000],
                "speed": self.DRIVE_SPEED,
            },
            callback=self.mock,
        )

        while self.mock.call_count != 4:
            pass
        self.assertEqual(
            self.mock.call_args.args[0]["error"], "Manipulator movement canceled"
        )

    def wait_for_callback(self):
        """Wait for callback to be called"""
        while not self.mock.called:
            pass
        self.mock.called = False



import atexit
import csv
import os.path
import time

# noinspection PyPackageRequirements
import socketio

from src.ephys_link import PositionalOutputData

# Parameters
MANIPULATOR_ID = 2
IP = "localhost"
PORT = 8080
HOME_POSITION = [10, 10, 10, 10]  # in mm

MOVEMENT_SPEED = 1
RESET_SPEED = 1000  # in um/s
TARGET_AXIS = 0  # 0 = X, 1 = Y, 2 = Z, 3 = Depth
TARGET_POS = 15  # in mm
FILE_NAME_EXT = "x1"

# Start time record
START_TIME = -1

# Setup server connection
sio = socketio.Client()
# noinspection HttpUrlsUsage
sio.connect(f"http://{IP}:{PORT}")

# Setup manipulator
sio.emit("register_manipulator", MANIPULATOR_ID)
sio.emit("bypass_calibration", MANIPULATOR_ID)
sio.emit(
    "set_can_write",
    {"manipulator_id": MANIPULATOR_ID, "can_write": True, "hours": 1},
)

# Setup output file and data dir
if not os.path.exists("data"):
    os.makedirs("data")
file = open(f"data/crash_test_{FILE_NAME_EXT}.csv", "a", newline="")
writer = csv.writer(file)
writer.writerow(["time (ms)", "x", "y", "z", "depth"])
file.flush()


# Define cleanup function
@atexit.register
def cleanup(_, __):
    print("Cleaning up...")
    file.flush()
    file.close()
    sio.disconnect()


# Define recording function
def record_position(data: PositionalOutputData):
    if data["error"] == "":
        writer.writerow([(time.time() * 1000) - START_TIME, *data["position"]])
        file.flush()
        sio.emit("get_pos", MANIPULATOR_ID, callback=record_position)


# Move to 0, 0, 0, 0
sio.emit(
    "goto_pos",
    {"manipulator_id": MANIPULATOR_ID, "pos": HOME_POSITION, "speed": RESET_SPEED},
)

# Set target location
target_location = HOME_POSITION
target_location[TARGET_AXIS] = TARGET_POS

# Update start time
START_TIME = time.time() * 1000

# Drive
sio.emit(
    "goto_pos",
    {"manipulator_id": MANIPULATOR_ID, "pos": target_location, "speed": MOVEMENT_SPEED},
)

# Record data
sio.emit("get_pos", MANIPULATOR_ID, callback=record_position)



from unittest import TestCase
from unittest.mock import Mock

# noinspection PyPackageRequirements
import socketio

from src.ephys_link import PositionalOutputData


class CalibrationTestCase(TestCase):
    """Tests calibration event"""

    def setUp(self):
        """Setup test case"""
        self.sio = socketio.Client()
        self.mock = Mock()

        self.sio.connect("http://localhost:8080")

    def test_calibrate_unregistered(self):
        """Test calibrate event with unregistered manipulator"""
        self.sio.emit("calibrate", 1, callback=self.mock)
        self.wait_for_callback()

        self.mock.assert_called_with("Manipulator not registered")

    def test_get_pos_uncalibrated(self):
        """Test get_pos event with uncalibrated manipulator"""
        self.sio.emit("register_manipulator", 1, callback=self.mock)
        self.wait_for_callback()
        self.sio.emit("get_pos", 1, callback=self.mock)
        self.wait_for_callback()
        self.mock.assert_called_with(
            PositionalOutputData([], "Manipulator not calibrated")
        )

    def test_move_uncalibrated(self):
        """Test move event with uncalibrated manipulator"""
        self.sio.emit("register_manipulator", 1, callback=self.mock)
        self.wait_for_callback()
        self.sio.emit(
            "set_can_write", {"manipulator_id": 1, "can_write": True, "hours": 1}
        )
        self.sio.emit(
            "goto_pos",
            {"manipulator_id": 1, "pos": [0, 0, 0, 0], "speed": 2000},
            callback=self.mock,
        )
        self.wait_for_callback()
        self.mock.assert_called_with(
            PositionalOutputData([], "Manipulator not calibrated")
        )

    def test_calibrate_registered(self):
        """Test calibrate event with registered manipulator"""
        self.sio.emit("register_manipulator", 1)
        self.sio.emit("register_manipulator", 2)

        self.sio.emit(
            "set_can_write", {"manipulator_id": 1, "can_write": True, "hours": 1}
        )
        self.sio.emit(
            "set_can_write", {"manipulator_id": 2, "can_write": True, "hours": 1}
        )

        self.sio.emit("calibrate", 1, callback=self.mock)
        self.sio.emit("calibrate", 2, callback=self.mock)

        while self.mock.call_count != 2:
            pass

        self.assertEqual("", self.mock.call_args.args[0])

    def tearDown(self) -> None:
        """Cleanup test case"""
        self.sio.disconnect()

    def wait_for_callback(self):
        """Wait for callback to be called"""
        while not self.mock.called:
            pass
        self.mock.called = False



# pylama:ignore=E402
from pythonnet import load

load("netfx")

# noinspection PyPackageRequirements
import clr

# noinspection PyUnresolvedReferences
clr.AddReference("../ephys_link/resources/NstMotorCtrl")

import random
import threading
import time

# noinspection PyUnresolvedReferences
from NstMotorCtrl import NstCtrlHostIntf

ctrl = NstCtrlHostIntf()

ctrl.Ports = "USB1,USB2"

print(ctrl.Initialize())

x = ctrl.GetAxis(0)
y = ctrl.GetAxis(1)

# Setup motion
x.SetCL_Enable(True)
x.SetCL_Speed(4000, 5000, 5)
y.SetCL_Enable(True)
y.SetCL_Speed(4000, 5000, 5)

target_pos = random.randint(0, 15000)
x.MoveAbsolute(target_pos)
y.MoveAbsolute(target_pos)
print("Target:", target_pos)

AT_TARGET_FLAG = 0x040000

done_event = threading.Event()


def check_done():
    x.QueryPosStatus()
    y.QueryPosStatus()
    while not x.CurStatus & AT_TARGET_FLAG and not y.CurStatus & AT_TARGET_FLAG:
        time.sleep(0.1)
        x.QueryPosStatus()
        y.QueryPosStatus()
    done_event.set()


threading.Thread(target=check_done).start()

done_event.wait()

x.QueryPosStatus()
y.QueryPosStatus()
print("Current pos x:", x.CurPosition)
print("Current pos y:", y.CurPosition)

ctrl.Disconnect()



# pylama:skip=1
import socket
from argparse import Namespace
from threading import Event, Thread
from tkinter import *
from tkinter import ttk

# GUI Variables
is_running = False
server_port: IntVar
platform_type: StringVar
new_scale_port: IntVar
e_stop_serial_port: StringVar
server_launch_button_text: StringVar


class GUI:
    """GUI definition for Ephys Link

    :param root: Root object of the Tk GUI
    :type root: Tk
    """

    def __init__(
        self,
        root: Tk,
        launch_func: callable,
        manipulator_stop_func: callable,
        poll_serial_func: callable,
        parsed_args: Namespace,
    ) -> None:
        """Setup and construction of the Tk GUI"""

        # Fields
        self._root = root
        self._launch_func = launch_func
        self._manipulator_stop_func = manipulator_stop_func
        self._poll_serial_func = poll_serial_func
        self._kill_serial_event = Event()
        self._parsed_args = parsed_args

        # Update GUI variables with defaults
        global server_port, platform_type, new_scale_port, e_stop_serial_port
        global server_launch_button_text
        server_port = IntVar(value=self._parsed_args.port)
        platform_type = StringVar(value=self._parsed_args.type)
        new_scale_port = IntVar(value=self._parsed_args.new_scale_port)
        e_stop_serial_port = StringVar(value=self._parsed_args.serial)
        server_launch_button_text = StringVar(value="Start Server")

        # Build GUI
        self.build_gui()

    def build_gui(self):
        """Build GUI"""

        self._root.title("Ephys Link")

        mainframe = ttk.Frame(self._root, padding=3)
        mainframe.grid(column=0, row=0, sticky="news")
        self._root.columnconfigure(0, weight=1)
        self._root.rowconfigure(0, weight=1)
        mainframe.columnconfigure(0, weight=1)
        mainframe.rowconfigure(0, weight=1)

        # Server serving settings

        server_serving_settings = ttk.LabelFrame(
            mainframe, text="Serving settings", padding=3
        )
        server_serving_settings.grid(column=0, row=0, sticky="news")

        # IP
        ttk.Label(server_serving_settings, text="IP:", anchor=E, justify=RIGHT).grid(
            column=0, row=0, sticky="we"
        )
        ttk.Label(
            server_serving_settings, text=socket.gethostbyname(socket.gethostname())
        ).grid(column=1, row=0, sticky="we")

        # Port
        ttk.Label(server_serving_settings, text="Port:", anchor=E, justify=RIGHT).grid(
            column=0, row=1, sticky="we"
        )
        ttk.Entry(
            server_serving_settings, textvariable=server_port, width=5, justify=CENTER
        ).grid(column=1, row=1, sticky="we")

        # ---

        # Platform type
        platform_type_settings = ttk.LabelFrame(
            mainframe, text="Platform Type", padding=3
        )
        platform_type_settings.grid(column=0, row=1, sticky="news")

        ttk.Radiobutton(
            platform_type_settings,
            text="Sensapex uMp",
            variable=platform_type,
            value="sensapex",
        ).grid(column=0, row=0, sticky="we")
        ttk.Radiobutton(
            platform_type_settings,
            text="New Scale",
            variable=platform_type,
            value="new_scale",
        ).grid(column=0, row=1, sticky="we")

        # ---

        # New Scale Settings
        new_scale_settings = ttk.LabelFrame(
            mainframe, text="New Scale settings", padding=3
        )
        new_scale_settings.grid(column=0, row=2, sticky="news")

        # Port
        ttk.Label(
            new_scale_settings, text="HTTP Server Port:", anchor=E, justify=RIGHT
        ).grid(column=0, row=1, sticky="we")
        ttk.Entry(
            new_scale_settings, textvariable=new_scale_port, width=5, justify=CENTER
        ).grid(column=1, row=1, sticky="we")

        # ---

        # Emergency Stop serial port
        e_stop_settings = ttk.LabelFrame(
            mainframe, text="Emergency Stop Settings", padding=3
        )
        e_stop_settings.grid(column=0, row=3, sticky="news")

        # Serial Port
        ttk.Label(e_stop_settings, text="Serial Port:", anchor=E, justify=RIGHT).grid(
            column=0, row=1, sticky="we"
        )
        ttk.Entry(
            e_stop_settings, textvariable=e_stop_serial_port, justify=CENTER
        ).grid(column=1, row=1, sticky="we")

        # Server start/stop button
        ttk.Button(
            mainframe,
            textvariable=server_launch_button_text,
            command=lambda: self.start_stop_server(not is_running),
        ).grid(column=0, row=4, columnspan=2, sticky="we")

    def start_stop_server(self, start: bool) -> None:
        """Start/stop server and update button text

        :param start: Whether to start or stop the server
        :type start: bool
        :return None
        """
        global server_port, server_launch_button_text, is_running
        is_running = not is_running
        if start:
            # Launch serial
            Thread(
                target=self._poll_serial_func,
                args=(
                    self._kill_serial_event,
                    e_stop_serial_port.get(),
                ),
                daemon=True,
            ).start()
            # Launch server
            Thread(
                target=self._launch_func,
                args=(platform_type.get(), server_port.get(), new_scale_port.get()),
                daemon=True,
            ).start()

            # Update UI
            server_launch_button_text.set("Close Server")
        else:
            # Stop serial
            self._kill_serial_event.set()

            # Stop manipulators
            self._manipulator_stop_func(0)

            # Close
            self._root.destroy()



"""Sensapex Manipulator class

Handles logic for calling Sensapex API functions. Also includes extra logic for safe
function calls, error handling, managing per-manipulator attributes, and returning the
appropriate callback parameters like in :mod:`ephys_link.sensapex_handler`.
"""

import asyncio
import threading

# noinspection PyPackageRequirements
import socketio
from sensapex import SensapexDevice

import ephys_link.common as com
from ephys_link.platform_manipulator import (
    HOURS_TO_SECONDS,
    MM_TO_UM,
    POSITION_POLL_DELAY,
    PlatformManipulator,
)


class SensapexManipulator(PlatformManipulator):
    """Representation of a single Sensapex manipulator

    :param device: A Sensapex device
    :type device: :class: `sensapex.SensapexDevice`
    """

    def __init__(self, device: SensapexDevice) -> None:
        """Construct a new Manipulator object

        :param device: A Sensapex device
        """
        super().__init__()
        self._device = device
        self._id = device.dev_id

    # Device functions
    def get_pos(self) -> com.PositionalOutputData:
        """Get the current position of the manipulator and convert it into mm

        :return: Callback parameters (position in (x, y, z, w) (or an empty array on
            error) in mm, error message)
        :rtype: :class:`ephys_link.common.PositionalOutputData`
        """
        try:
            position = [axis / MM_TO_UM for axis in self._device.get_pos(1)]
            com.dprint(f"[SUCCESS]\t Got position of manipulator {self._id}\n")
            return com.PositionalOutputData(position, "")
        except Exception as e:
            print(f"[ERROR]\t\t Getting position of manipulator {self._id}")
            print(f"{e}\n")
            return com.PositionalOutputData([], "Error getting position")

    async def goto_pos(
        self, position: list[float], speed: float
    ) -> com.PositionalOutputData:
        """Move manipulator to position

        :param position: The position to move to in mm
        :type position: list[float]
        :param speed: The speed to move at (in mm/s)
        :type speed: float
        :return: Callback parameters (position in (x, y, z, w) (or an empty array on
            error), error message)
        :rtype: :class:`ephys_link.common.PositionalOutputData`
        """
        # Check if able to write
        if not self._can_write:
            print(f"[ERROR]\t\t Manipulator {self._id} movement " f"canceled")
            return com.PositionalOutputData([], "Manipulator " "movement canceled")

        # Stop current movement
        if self._is_moving:
            self._device.stop()
            self._is_moving = False

        try:
            target_position_um = [axis * MM_TO_UM for axis in position]

            # Restrict target position to just depth-axis if inside brain
            if self._inside_brain:
                d_axis = target_position_um[3]
                target_position_um = self._device.get_pos()
                target_position_um[3] = d_axis

            # Mark movement as started
            self._is_moving = True

            # Send move command
            movement = self._device.goto_pos(target_position_um, speed * MM_TO_UM)

            # Wait for movement to finish
            while not movement.finished:
                await asyncio.sleep(POSITION_POLL_DELAY)

            # Get position
            manipulator_final_position = self.get_pos()["position"]

            # Mark movement as finished
            self._is_moving = False

            com.dprint(
                f"[SUCCESS]\t Moved manipulator {self._id} to position"
                f" {manipulator_final_position}\n"
            )
            return com.PositionalOutputData(manipulator_final_position, "")
        except Exception as e:
            print(
                f"[ERROR]\t\t Moving manipulator {self._id} to position" f" {position}"
            )
            print(f"{e}\n")
            return com.PositionalOutputData([], "Error moving manipulator")

    async def drive_to_depth(
        self, depth: float, speed: int
    ) -> com.DriveToDepthOutputData:
        """Drive the manipulator to a certain depth

        :param depth: The depth to drive to in mm
        :type depth: float
        :param speed: The speed to drive at in mm/s
        :type speed: int
        :return: Callback parameters (depth (or 0 on error), error message)
        :rtype: :class:`ephys_link.common.DriveToDepthOutputData`
        """
        # Get position before this movement
        target_pos = self.get_pos()["position"]

        target_pos[3] = depth
        movement_result = await self.goto_pos(target_pos, speed)

        if movement_result["error"] == "":
            # Return depth on success
            return com.DriveToDepthOutputData(movement_result["position"][3], "")
        else:
            # Return 0 and error message on failure
            return com.DriveToDepthOutputData(0, "Error driving " "manipulator")

    def set_inside_brain(self, inside: bool) -> None:
        """Set if the manipulator is inside the brain

        Used to signal that the brain should move at :const:`INSIDE_BRAIN_SPEED_LIMIT`

        :param inside: True if the manipulator is inside the brain, False otherwise
        :type inside: bool
        :return: None
        """
        self._inside_brain = inside

    def get_can_write(self) -> bool:
        """Return if the manipulator can move

        :return: True if the manipulator can move, False otherwise
        :rtype: bool
        """
        return self._can_write

    def set_can_write(
        self, can_write: bool, hours: float, sio: socketio.AsyncServer
    ) -> None:
        """Set if the manipulator can move

        :param can_write: True if the manipulator can move, False otherwise
        :type can_write: bool
        :param hours: The number of hours to allow the manipulator to move (0 =
            forever)
        :type hours: float
        :param sio: SocketIO object from server to emit reset event
        :type sio: :class:`socketio.AsyncServer`
        :return: None
        """
        self._can_write = can_write

        if can_write and hours > 0:
            if self._reset_timer:
                self._reset_timer.cancel()
            self._reset_timer = threading.Timer(
                hours * HOURS_TO_SECONDS, self.reset_can_write, [sio]
            )
            self._reset_timer.start()

    def reset_can_write(self, sio: socketio.AsyncServer) -> None:
        """Reset the :attr:`can_write` flag

        :param sio: SocketIO object from server to emit reset event
        :type sio: :class:`socketio.AsyncServer`
        :return: None
        """
        self._can_write = False
        asyncio.run(sio.emit("write_disabled", self._id))

    # Calibration
    def call_calibrate(self) -> None:
        """Calibrate the manipulator

        :return: None
        """
        self._device.calibrate_zero_position()

    def get_calibrated(self) -> bool:
        """Return the calibration state of the manipulator.

        :return: True if the manipulator is calibrated, False otherwise
        :rtype: bool
        """
        return self._calibrated

    def set_calibrated(self) -> None:
        """Set the manipulator to calibrated

        :return: None
        """
        self._calibrated = True

    def stop(self) -> None:
        """Stop the manipulator

        :return: None
        """
        self._can_write = False
        self._device.stop()
        com.dprint(f"[SUCCESS]\t Stopped manipulator {self._id}")
